# pylint: disable=no-self-use

import numpy as np
import pytest

from PartSegCore.image_operations import RadiusType
from PartSegCore.mask_create import MaskProperty, calculate_mask, fill_2d_holes_in_mask, fill_holes_in_mask
from PartSegImage import Image


class TestMaskProperty:
    def test_compare(self):
        ob1 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=True,
        )
        ob2 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=True,
            reversed_mask=False,
        )
        ob3 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=True,
            reversed_mask=True,
        )
        assert ob1 == ob2
        assert ob1 != ob3
        assert ob2 != ob3

    def test_str(self):
        dilate_text = "dilate radius"
        hole_text = "max holes size"
        text = str(
            MaskProperty(
                dilate=RadiusType.NO,
                dilate_radius=0,
                fill_holes=RadiusType.NO,
                max_holes_size=0,
                save_components=False,
                clip_to_mask=True,
            )
        )
        assert dilate_text not in text
        assert hole_text not in text
        text = str(
            MaskProperty(
                dilate=RadiusType.R2D,
                dilate_radius=0,
                fill_holes=RadiusType.NO,
                max_holes_size=0,
                save_components=False,
                clip_to_mask=True,
            )
        )
        assert dilate_text in text
        assert hole_text not in text
        text = str(
            MaskProperty(
                dilate=RadiusType.NO,
                dilate_radius=0,
                fill_holes=RadiusType.R2D,
                max_holes_size=0,
                save_components=False,
                clip_to_mask=True,
            )
        )
        assert dilate_text not in text
        assert hole_text in text
        text = str(
            MaskProperty(
                dilate=RadiusType.R3D,
                dilate_radius=0,
                fill_holes=RadiusType.R2D,
                max_holes_size=0,
                save_components=False,
                clip_to_mask=True,
            )
        )
        assert dilate_text in text
        assert hole_text in text

    def test_simple_mask(self):
        assert MaskProperty.simple_mask() == MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )


class TestFillHoles:
    def test_fill_2d(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[:, 4:6, 4:6] = 0
        assert np.all(mask == fill_2d_holes_in_mask(mask2, -1))

    def test_fill_2d_min_size(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[:, 4:6, 4:6] = 0
        mask3 = fill_2d_holes_in_mask(mask2, 1)
        assert np.all(mask3 == mask2)
        mask3 = fill_2d_holes_in_mask(mask2, 3)
        assert np.all(mask3 == mask2)
        mask3 = fill_2d_holes_in_mask(mask2, 4)
        assert np.all(mask3 == mask)
        mask3 = fill_2d_holes_in_mask(mask2, 15)
        assert np.all(mask3 == mask)

    def test_fill_3d(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[:, 4:6, 4:6] = 0
        assert not np.all(mask == fill_holes_in_mask(mask2, -1))
        mask2 = np.copy(mask)
        mask2[4:6, 4:6, 4:6] = 0
        assert np.all(mask == fill_holes_in_mask(mask2, -1))

    def test_fill_3d_min_size(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[4:6, 4:6, 4:6] = 0
        assert not np.all(mask == fill_holes_in_mask(mask2, 6))
        assert np.all(mask == fill_holes_in_mask(mask2, 8))
        assert np.all(mask == fill_holes_in_mask(mask2, 16))

    def test_on_2d_data(self):
        mask = np.zeros((10, 10), dtype=np.uint8)
        mask[2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[4:6, 4:6] = 0
        assert np.all(mask == fill_holes_in_mask(mask2, -1))
        assert np.all(mask == fill_2d_holes_in_mask(mask2, -1))


class TestCalculateMask:
    def test_single(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask2 = calculate_mask(mp, mask, None, (1, 1, 1))
        assert np.all(mask2 == mask)
        mask3 = np.copy(mask)
        mask3[mask > 0] = 2
        mask2 = calculate_mask(mp, mask, None, (1, 1, 1))
        assert np.all(mask2 == mask)

    def test_dilate(self):
        mask = np.zeros((1, 10, 10, 10), dtype=np.uint8)
        mask[0, 2:8, 2:8, 2:8] = 1
        mp = MaskProperty(
            dilate=RadiusType.R3D,
            dilate_radius=-1,
            fill_holes=RadiusType.NO,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask2 = calculate_mask(mp, mask, None, (1, 1, 1))
        mask3 = np.zeros((1, 10, 10, 10), dtype=np.uint8)
        mask3[0, 3:7, 3:7, 3:7] = 1
        assert np.all(mask2 == mask3)
        mp.dilate = RadiusType.R2D
        mask2 = calculate_mask(mp, mask, None, (1, 1, 1))
        mask3 = np.zeros((1, 10, 10, 10), dtype=np.uint8)
        mask3[0, 2:8, 3:7, 3:7] = 1
        assert np.all(mask2 == mask3)

    def test_fil_holes_3d(self):
        mask = np.zeros((1, 10, 10, 10), dtype=np.uint8)
        mask[0, 2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[0, 4:6, 4:6, 4:6] = 0
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R3D,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask)
        mp.max_holes_size = 7
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask2)
        mp.max_holes_size = 8
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask)

    def test_fil_holes_3d_torus(self):
        mask = np.zeros((1, 10, 10, 10), dtype=np.uint8)
        mask[0, 2:8, 2:8, 2:8] = 1
        mask[0, :, 4:6, 4:6] = 0
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R3D,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask2 = calculate_mask(mp, mask, None, (1, 1, 1))
        assert np.all(mask2 == mask)

    def test_fil_holes_2d(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[:, 4:6, 4:6] = 0
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R2D,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask)
        mp.max_holes_size = 3
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask2)
        mp.max_holes_size = 4
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask)

    def test_save_components(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[2:8, 2:8, 2:8] = 1
        mask2 = np.copy(mask)
        mask2[2:8, 4:6, 4:6] = 2
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask)
        mp.save_components = True
        mask3 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask3 == mask2)
        mask4 = np.copy(mask2)
        mask4[mask4 == 2] = 3
        mask3 = calculate_mask(mp, mask4, None, (1, 1, 1))
        assert np.all(mask3 == mask4)

    def test_save_component_fill_holes(self):
        mask = np.zeros((12, 12, 12), dtype=np.uint8)
        mask[2:7, 2:-2, 2:-2] = 1
        mask[7:-2, 2:-2, 2:-2] = 2
        mask[4:-4, 4:-4, 4:-4] = 3
        mask2 = np.copy(mask)
        mask2[5, 5, 5] = 0
        mask2[3, 3, 3] = 0
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R2D,
            max_holes_size=-1,
            save_components=True,
            clip_to_mask=True,
        )
        mp2 = mp.copy(update={"fill_holes": RadiusType.R3D})
        mask1 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)
        mask1 = calculate_mask(mp2, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)
        mask[2:-2, 4:-4, 4:-4] = 3
        mask2 = np.copy(mask)
        mask2[5, 5, 5] = 0
        mask2[3, 3, 3] = 0
        mask1 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)
        mask1 = calculate_mask(mp2, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)
        mask[2:7, 2:-2, 2:-2] = 1
        mask[4:-4, 4:-4, 4:-4] = 2
        mask2 = np.copy(mask)
        mask2[5, 5, 5] = 0
        mask2[3, 3, 3] = 0
        mask1 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)
        mask1 = calculate_mask(mp2, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)

    @pytest.mark.xfail(reason="problem with alone pixels")
    def test_save_component_fill_holes_problematic(self):
        mask = np.zeros((12, 12, 12), dtype=np.uint8)
        mask[2:7, 2:-2, 2:-2] = 3
        mask[7:-2, 2:-2, 2:-2] = 2
        mask[4:-4, 4:-4, 4:-4] = 1
        mask2 = np.copy(mask)
        mask2[5, 5, 5] = 0
        mask2[3, 3, 3] = 0
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R2D,
            max_holes_size=-1,
            save_components=True,
            clip_to_mask=True,
        )
        mask1 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)
        mask1 = calculate_mask(mp, mask2, None, (1, 1, 1))
        assert np.all(mask == mask1)

    def test_reverse(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[3:7, 3:7, 3:7] = 1
        mask2 = np.ones((10, 10, 10), dtype=np.uint8)
        mask2[mask > 0] = 0
        mask3 = calculate_mask(
            MaskProperty(
                dilate=RadiusType.NO,
                dilate_radius=0,
                fill_holes=RadiusType.R2D,
                max_holes_size=-1,
                save_components=False,
                clip_to_mask=True,
                reversed_mask=True,
            ),
            mask,
            None,
            (1, 1, 1),
        )
        assert np.all(mask3 == mask2)

    def test_clip(self):
        mask = np.zeros((10, 10, 10), dtype=np.uint8)
        mask[3:7, 3:7, 3:7] = 1
        mp = MaskProperty(
            dilate=RadiusType.R3D,
            dilate_radius=1,
            fill_holes=RadiusType.NO,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=True,
        )
        mask2 = calculate_mask(mp, mask, mask, (1, 1, 1))
        assert np.all(mask == mask2)
        mask2[:, 4:6, 4:6] = 0
        mask3 = calculate_mask(mp, mask, mask2, (1, 1, 1))
        assert np.all(mask3 == mask2)
        mask2[:] = 0
        mask2[4:6, 4:6, 4:6] = 1
        mask3 = calculate_mask(
            mp.copy(update={"dilate": RadiusType.NO, "dilate_radius": 0, "reversed_mask": True}), mask2, mask, (1, 1, 1)
        )
        mask[mask2 == 1] = 0
        assert np.all(mask3 == mask)

    def test_chose_components(self):
        mask = np.zeros((12, 12, 12), dtype=np.uint8)
        mask[2:7, 2:-2, 2:-2] = 3
        mask[7:-2, 2:-2, 2:-2] = 2
        mask[4:-4, 4:-4, 4:-4] = 1
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=-1,
            save_components=True,
            clip_to_mask=True,
        )
        mask1 = calculate_mask(mp, mask, None, (1, 1, 1), [1, 3])
        assert np.all(np.unique(mask1) == [0, 1, 3])

    @pytest.mark.parametrize("dilate", [RadiusType.NO, RadiusType.R2D, RadiusType.R3D])
    @pytest.mark.parametrize("radius", [-1, 0, 1, 3])
    @pytest.mark.parametrize("fill_holes", [RadiusType.NO, RadiusType.R2D, RadiusType.R3D])
    @pytest.mark.parametrize("max_holes_size", [-1, 0, 100])
    @pytest.mark.parametrize("save_components", [True, False])
    @pytest.mark.parametrize("clip_to_mask", [True, False])
    @pytest.mark.parametrize("reversed_mask", [True, False])
    @pytest.mark.parametrize("old_mask", [True, False])
    def test_mask_property_combinations(
        self, dilate, radius, fill_holes, max_holes_size, save_components, clip_to_mask, reversed_mask, old_mask
    ):
        mask = np.zeros((1, 6, 6, 15), dtype=np.uint8)
        im = Image(data=mask.copy(), spacing=(3, 1, 1), file_path="", axes_order="TZYX")
        mask[:, 1:-1, 1:-1, 2:5] = 1
        mask[:, 2:-2, 2:-2, 3:4] = 0
        mask[:, 1:-1, 1:-1, 6:9] = 2
        mask[:, 2:-2, 2:-2, 7:8] = 0
        mask[:, 1:-1, 1:-1, 10:13] = 3
        mask[:, 2:-2, 2:-2, 11:12] = 0
        mask = im.fit_mask_to_image(mask)
        assert np.all(np.unique(mask.flat) == [0, 1, 2, 3])
        _old_mask = np.zeros(mask.shape, dtype=mask.dtype) if old_mask else None
        mp = MaskProperty(
            dilate=dilate,
            dilate_radius=radius,
            fill_holes=fill_holes,
            max_holes_size=max_holes_size,
            save_components=save_components,
            clip_to_mask=clip_to_mask,
            reversed_mask=reversed_mask,
        )
        mask1 = calculate_mask(mp, mask, _old_mask, im.spacing, time_axis=im.time_pos)
        assert mask1.shape == mask.shape

    @pytest.mark.parametrize("time", [1, 2, 4])
    def test_time_axis(self, time):
        mask = np.zeros((time, 6, 6, 6), dtype=np.uint8)
        mask[:, 1:-1, 1:-1, 2:5] = 1
        mask[:, 2:-2, 2:-2, 3:4] = 0
        mp = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=-1,
            save_components=False,
            clip_to_mask=False,
        )
        mask1 = calculate_mask(mp, mask, None, (1, 1, 1))
        assert mask1.shape == mask.shape


# TODO add test with touching boundaries.



# pylint: disable=no-self-use

import operator
from abc import ABC
from copy import deepcopy
from typing import Union

import numpy as np
import pytest
from pydantic import BaseModel

from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis.algorithm_description import AnalysisAlgorithmSelection
from PartSegCore.analysis.analysis_utils import SegmentationPipeline, SegmentationPipelineElement
from PartSegCore.analysis.calculate_pipeline import calculate_pipeline
from PartSegCore.convex_fill import _convex_fill, convex_fill
from PartSegCore.image_operations import RadiusType
from PartSegCore.mask_create import MaskProperty, calculate_mask
from PartSegCore.roi_info import BoundInfo, ROIInfo
from PartSegCore.segmentation import ROIExtractionAlgorithm, algorithm_base
from PartSegCore.segmentation import restartable_segmentation_algorithms as sa
from PartSegCore.segmentation.noise_filtering import NoiseFilterSelection
from PartSegCore.segmentation.watershed import WatershedSelection
from PartSegImage import Image


def get_two_parts_array():
    data = np.zeros((1, 50, 100, 100), dtype=np.uint16)
    data[0, 10:40, 10:40, 10:90] = 50
    data[0, 10:40, 50:90, 10:90] = 50
    data[0, 15:35, 15:35, 15:85] = 70
    data[0, 15:35, 55:85, 15:85] = 70
    data[0, 10:40, 40:50, 10:90] = 40
    return data


def get_two_parts():
    return Image(get_two_parts_array(), spacing=(100, 50, 50), file_path="", axes_order="TZYX")


def get_two_parts_reversed():
    data = get_two_parts_array()
    data = 100 - data
    return Image(data, spacing=(100, 50, 50), file_path="", axes_order="TZYX")


def get_multiple_part_array(part_num):
    data = np.zeros((1, 20, 40, 40 * part_num), dtype=np.uint8)
    data[0, 4:16, 8:32, 8 : 40 * part_num - 8] = 40
    for i in range(part_num):
        data[0, 5:15, 10:30, 40 * i + 10 : 40 * i + 30] = 50
        data[0, 7:13, 15:25, 40 * i + 15 : 40 * i + 25] = 70
    return data


def get_multiple_part(part_num):
    return Image(get_multiple_part_array(part_num), spacing=(100, 50, 50), file_path="", axes_order="TZYX")


def get_multiple_part_reversed(part_num):
    data = 100 - get_multiple_part_array(part_num)
    return Image(data, spacing=(100, 50, 50), file_path="", axes_order="TZYX")


def get_two_parts_side():
    data = get_two_parts_array()
    data[0, 25, 40:45, 50] = 49
    data[0, 25, 45:50, 51] = 49
    return Image(data, spacing=(100, 50, 50), file_path="", axes_order="TZYX")


def get_two_parts_side_reversed():
    data = get_two_parts_array()
    data[0, 25, 40:45, 50] = 49
    data[0, 25, 45:50, 51] = 49
    data = 100 - data
    return Image(data, spacing=(100, 50, 50), file_path="", axes_order="TZYX")


def empty(_s: str, _i: int):
    """mock function for callback"""


@pytest.mark.parametrize("algorithm_name", AnalysisAlgorithmSelection.__register__.keys())
def test_base_parameters(algorithm_name):
    algorithm_class = AnalysisAlgorithmSelection[algorithm_name]
    assert algorithm_class.get_name() == algorithm_name
    algorithm_class: type[ROIExtractionAlgorithm]
    obj = algorithm_class()
    values = algorithm_class.get_default_values()
    obj.set_parameters(values)
    parameters = obj.get_segmentation_profile()
    assert parameters.algorithm == algorithm_name
    assert parameters.values == values


class BaseThreshold:
    def check_result(self, result, sizes, op, parameters):
        assert result.roi.max() == len(sizes)
        assert np.all(op(np.bincount(result.roi.flat)[1:], np.array(sizes)))
        assert result.parameters.values == parameters
        assert result.parameters.algorithm == self.get_algorithm_class().get_name()

    def get_parameters(self) -> BaseModel:
        if hasattr(self, "parameters") and isinstance(self.parameters, (dict, BaseModel)):
            return deepcopy(self.parameters)
        raise NotImplementedError

    def get_shift(self):
        if hasattr(self, "shift"):
            return deepcopy(self.shift)
        raise NotImplementedError

    @staticmethod
    def get_base_object():
        raise NotImplementedError

    @staticmethod
    def get_side_object():
        raise NotImplementedError

    def get_algorithm_class(self) -> type[ROIExtractionAlgorithm]:
        raise NotImplementedError


class BaseOneThreshold(BaseThreshold, ABC):  # pylint: disable=W0223
    def test_simple(self):
        image = self.get_base_object()
        alg: ROIExtractionAlgorithm = self.get_algorithm_class()()
        parameters = self.get_parameters()
        alg.set_image(image)
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        self.check_result(result, [96000, 72000], operator.eq, parameters)

        parameters.threshold.values.threshold += self.get_shift()
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        self.check_result(result, [192000], operator.eq, parameters)

    def test_side_connection(self):
        image = self.get_side_object()
        alg: ROIExtractionAlgorithm = self.get_algorithm_class()()
        parameters = self.get_parameters()
        parameters.side_connection = True
        alg.set_image(image)
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        self.check_result(result, [96000 + 5, 72000 + 5], operator.eq, parameters)

        parameters.side_connection = False
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        self.check_result(result, [96000 + 5 + 72000 + 5], operator.eq, parameters)


class TestLowerThreshold(BaseOneThreshold):
    parameters = sa.LowerThresholdAlgorithm.__argument_class__(
        channel=0,
        minimum_size=30000,
        threshold={"name": "Manual", "values": {"threshold": 45}},
        noise_filtering={"name": "None", "values": {}},
        side_connection=False,
    )
    shift = -6

    @staticmethod
    def get_base_object():
        return get_two_parts()

    @staticmethod
    def get_side_object():
        return get_two_parts_side()

    def get_algorithm_class(self) -> type[ROIExtractionAlgorithm]:
        return sa.LowerThresholdAlgorithm


class TestUpperThreshold(BaseOneThreshold):
    parameters = sa.UpperThresholdAlgorithm.__argument_class__(
        channel=0,
        minimum_size=30000,
        threshold={"name": "Manual", "values": {"threshold": 55}},
        noise_filtering={"name": "None", "values": {}},
        side_connection=False,
    )
    shift = 6

    @staticmethod
    def get_base_object():
        return get_two_parts_reversed()

    @staticmethod
    def get_side_object():
        return get_two_parts_side_reversed()

    def get_algorithm_class(self) -> type[ROIExtractionAlgorithm]:
        return sa.UpperThresholdAlgorithm


class TestRangeThresholdAlgorithm:
    def test_simple(self):
        image = get_two_parts()
        alg = sa.RangeThresholdAlgorithm()
        parameters = sa.RangeThresholdAlgorithm.__argument_class__(
            threshold={
                "name": "Range",
                "values": {
                    "base_threshold": {
                        "name": "Manual",
                        "values": {
                            "threshold": 45,
                        },
                    },
                    "core_threshold": {
                        "name": "Manual",
                        "values": {
                            "threshold": 60,
                        },
                    },
                },
            },
            channel=0,
            minimum_size=8000,
            noise_filtering={"name": "None", "values": {}},
            side_connection=False,
        )
        alg.set_parameters(parameters)
        alg.set_image(image)
        result = alg.calculation_run(empty)
        assert np.max(result.roi) == 2
        assert np.all(
            np.bincount(result.roi.flat)[1:] == np.array([30 * 40 * 80 - 20 * 30 * 70, 30 * 30 * 80 - 20 * 20 * 70])
        )
        assert result.parameters.values == parameters
        assert result.parameters.algorithm == alg.get_name()

        parameters.threshold.values.base_threshold.values.threshold -= 6
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        assert np.max(result.roi) == 1
        assert np.bincount(result.roi.flat)[1] == 30 * 80 * 80 - 20 * 50 * 70
        assert result.parameters.values == parameters
        assert result.parameters.algorithm == alg.get_name()

    def test_side_connection(self):
        image = get_two_parts_side()
        alg = sa.RangeThresholdAlgorithm()
        parameters = sa.RangeThresholdAlgorithm.__argument_class__(
            threshold={
                "name": "Range",
                "values": {
                    "base_threshold": {
                        "name": "Manual",
                        "values": {
                            "threshold": 45,
                        },
                    },
                    "core_threshold": {
                        "name": "Manual",
                        "values": {
                            "threshold": 60,
                        },
                    },
                },
            },
            channel=0,
            minimum_size=8000,
            noise_filtering={"name": "None", "values": {}},
            side_connection=True,
        )
        alg.set_parameters(parameters)
        alg.set_image(image)
        result = alg.calculation_run(empty)
        assert np.max(result.roi) == 2
        assert np.all(
            np.bincount(result.roi.flat)[1:]
            == np.array([30 * 40 * 80 - 20 * 30 * 70 + 5, 30 * 30 * 80 - 20 * 20 * 70 + 5])
        )
        assert result.parameters.values == parameters
        assert result.parameters.algorithm == alg.get_name()

        parameters.side_connection = False
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        assert np.max(result.roi) == 1
        assert np.bincount(result.roi.flat)[1] == 30 * 70 * 80 - 20 * 50 * 70 + 10
        assert result.parameters.values == parameters
        assert result.parameters.algorithm == alg.get_name()


class BaseFlowThreshold(BaseThreshold, ABC):  # pylint: disable=W0223
    @pytest.mark.parametrize("sprawl_algorithm_name", WatershedSelection.__register__.keys())
    @pytest.mark.parametrize("compare_op", [operator.eq, operator.ge])
    @pytest.mark.parametrize("components", [2, *list(range(3, 15, 2))])
    def test_multiple(self, sprawl_algorithm_name, compare_op, components):
        alg = self.get_algorithm_class()()
        parameters = self.get_parameters()
        image = self.get_multiple_part(components)
        alg.set_image(image)
        sprawl_algorithm = WatershedSelection[sprawl_algorithm_name]
        parameters.flow_type = WatershedSelection(
            name=sprawl_algorithm_name, values=sprawl_algorithm.get_default_values()
        )
        if compare_op(1, 0):
            parameters.threshold.values.base_threshold.values.threshold += self.get_shift()
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        self.check_result(result, [4000] * components, compare_op, parameters)

    @pytest.mark.parametrize("algorithm_name", WatershedSelection.__register__.keys())
    def test_side_connection(self, algorithm_name):
        image = self.get_side_object()
        alg = self.get_algorithm_class()()
        parameters = self.get_parameters()
        parameters.side_connection = True
        alg.set_image(image)
        val = WatershedSelection[algorithm_name]
        parameters.flow_type = WatershedSelection(name=algorithm_name, values=val.get_default_values())
        alg.set_parameters(parameters)
        result = alg.calculation_run(empty)
        self.check_result(result, [96000 + 5, 72000 + 5], operator.eq, parameters)

    def get_multiple_part(self, parts_num):
        raise NotImplementedError


class TestLowerThresholdFlow(BaseFlowThreshold):
    parameters = sa.LowerThresholdFlowAlgorithm.__argument_class__(
        channel=0,
        minimum_size=30,
        threshold={
            "name": "Base/Core",
            "values": {
                "core_threshold": {"name": "Manual", "values": {"threshold": 55}},
                "base_threshold": {"name": "Manual", "values": {"threshold": 45}},
            },
        },
        noise_filtering={"name": "None", "values": {}},
        side_connection=False,
        flow_type={"name": "Euclidean", "values": {}},
    )
    shift = -6
    get_base_object = staticmethod(get_two_parts)
    get_side_object = staticmethod(get_two_parts_side)
    get_multiple_part = staticmethod(get_multiple_part)

    def get_algorithm_class(self) -> type[ROIExtractionAlgorithm]:
        return sa.LowerThresholdFlowAlgorithm


class TestUpperThresholdFlow(BaseFlowThreshold):
    parameters = sa.UpperThresholdFlowAlgorithm.__argument_class__(
        channel=0,
        minimum_size=30,
        threshold={
            "name": "Base/Core",
            "values": {
                "core_threshold": {"name": "Manual", "values": {"threshold": 45}},
                "base_threshold": {"name": "Manual", "values": {"threshold": 55}},
            },
        },
        noise_filtering={"name": "None", "values": {}},
        side_connection=False,
        flow_type={"name": "Euclidean", "values": {}},
    )
    shift = 6
    get_base_object = staticmethod(get_two_parts_reversed)
    get_side_object = staticmethod(get_two_parts_side_reversed)
    get_multiple_part = staticmethod(get_multiple_part_reversed)

    def get_algorithm_class(self) -> type[ROIExtractionAlgorithm]:
        return sa.UpperThresholdFlowAlgorithm


class TestMaskCreate:
    def test_simple_mask(self):
        mask_array = np.zeros((10, 20, 20), dtype=np.uint8)
        mask_array[3:7, 6:14, 6:14] = 1
        prop = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )
        new_mask = calculate_mask(prop, mask_array, None, (1, 1, 1))
        assert np.all(new_mask == mask_array)
        mask_array2 = np.copy(mask_array)
        mask_array2[4:6, 8:12, 8:12] = 2
        new_mask = calculate_mask(prop, mask_array2, None, (1, 1, 1))
        assert np.all(new_mask == mask_array)
        prop2 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=True,
            clip_to_mask=False,
        )
        new_mask = calculate_mask(prop2, mask_array2, None, (1, 1, 1))
        assert np.all(new_mask == mask_array2)

    def test_fill_holes(self):
        mask_base_array = np.zeros((1, 20, 30, 30), dtype=np.uint8)
        mask_base_array[:, 4:16, 8:22, 8:22] = 1
        mask1_array = np.copy(mask_base_array)
        mask1_array[:, 4:16, 10:15, 10:15] = 0
        prop = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R2D,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )
        new_mask = calculate_mask(prop, mask1_array, None, (1, 1, 1))
        assert np.all(mask_base_array == new_mask)

        prop = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R3D,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )
        new_mask = calculate_mask(prop, mask1_array, None, (1, 1, 1))
        assert np.all(mask1_array == new_mask)

        mask2_array = np.copy(mask1_array)
        mask2_array[:, 5:15, 10:15, 17:20] = 0
        new_mask = calculate_mask(prop, mask2_array, None, (1, 1, 1))
        assert np.all(mask1_array == new_mask)

    def test_fill_holes_components(self):
        mask_base_array = np.zeros((20, 30, 30), dtype=np.uint8)
        mask_base_array[4:16, 6:15, 6:24] = 1
        mask_base_array[4:16, 15:24, 6:24] = 2
        res_mask1 = (mask_base_array > 0).astype(np.uint8)
        res_mask2 = np.copy(mask_base_array)
        mask_base_array[6:14, 8:12, 8:22] = 0
        mask_base_array[6:14, 18:22, 8:22] = 0
        prop1 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R3D,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )
        prop2 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R3D,
            max_holes_size=0,
            save_components=True,
            clip_to_mask=False,
        )
        new_mask = calculate_mask(prop1, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == res_mask1)
        new_mask = calculate_mask(prop2, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == res_mask2)

        mask_base_array[6:14, 14:16, 8:22] = 0
        res_mask2[6:14, 14:16, 8:22] = 0
        new_mask = calculate_mask(prop1, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == res_mask1)
        new_mask = calculate_mask(prop2, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == res_mask2)

    def test_fill_holes_size(self):
        mask_base_array = np.zeros((1, 20, 20, 40), dtype=np.uint8)
        mask_base_array[0, 2:18, 2:18, 4:36] = 1
        mask_base_array[0, 4:16, 4:16, 6:18] = 0
        mask1_array = np.copy(mask_base_array)
        mask1_array[0, 6:14, 6:14, 24:32] = 0

        prop1 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R2D,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=False,
        )
        prop2 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=0,
            fill_holes=RadiusType.R3D,
            max_holes_size=530,
            save_components=True,
            clip_to_mask=False,
        )

        new_mask = calculate_mask(prop1, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == mask_base_array)
        new_mask = calculate_mask(prop2, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == mask_base_array)

    @pytest.mark.parametrize("radius_type", [RadiusType.R2D, RadiusType.R3D])
    @pytest.mark.parametrize("radius", [1, -1])
    @pytest.mark.parametrize("time", [1, 2, 5])
    def test_dilate(self, radius_type, radius, time):
        mask_base_array = np.zeros((time, 30, 30, 30), dtype=np.uint8)
        mask_base_array[:, 10:20, 10:20, 10:20] = 1
        prop1 = MaskProperty(
            dilate=radius_type,
            dilate_radius=radius,
            fill_holes=RadiusType.NO,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=False,
        )
        res_array1 = np.zeros((1, 30, 30, 30), dtype=np.uint8)
        slices: list[Union[int, slice]] = [slice(None)] * 4
        for i in range(1, 4):
            slices[i] = slice(10 - radius, 20 + radius)
        if radius_type == RadiusType.R2D:
            slices[1] = slice(10, 20)
        res_array1[tuple(slices)] = 1
        if radius_type == RadiusType.R3D:
            res_array1[:, (9, 9, 9, 9), (9, 9, 20, 20), (9, 20, 20, 9)] = 0
            res_array1[:, (20, 20, 20, 20), (9, 9, 20, 20), (9, 20, 20, 9)] = 0
        new_mask = calculate_mask(prop1, mask_base_array, None, (1, 1, 1))
        assert np.all(new_mask == res_array1)

    @pytest.mark.parametrize("radius", [-1, -2, -3])
    def test_dilate_spacing_negative(self, radius):
        mask_base_array = np.zeros((1, 30, 30, 30), dtype=np.uint8)
        mask_base_array[:, 10:20, 5:25, 5:25] = 1
        res_array1 = np.zeros((1, 30, 30, 30), dtype=np.uint8)
        s = slice(10, 20) if radius == -1 else slice(11, 19)
        res_array1[:, s, 5 - radius : 25 + radius, 5 - radius : 25 + radius] = 1

        prop1 = MaskProperty(
            dilate=RadiusType.R3D,
            dilate_radius=radius,
            fill_holes=RadiusType.NO,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=False,
        )
        new_mask = calculate_mask(prop1, mask_base_array, None, (3, 1, 1))
        assert np.all(new_mask == res_array1)

    def test_dilate_spacing_positive(self):
        mask_base_array = np.zeros((1, 30, 30, 30), dtype=np.uint8)
        mask_base_array[:, 10:20, 10:20, 10:20] = 1
        prop1 = MaskProperty(
            dilate=RadiusType.R3D,
            dilate_radius=1,
            fill_holes=RadiusType.NO,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=False,
        )
        prop2 = MaskProperty(
            dilate=RadiusType.R3D,
            dilate_radius=2,
            fill_holes=RadiusType.NO,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=False,
        )
        prop3 = MaskProperty(
            dilate=RadiusType.R3D,
            dilate_radius=3,
            fill_holes=RadiusType.NO,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=False,
        )
        res_array1 = np.zeros((30, 30, 30), dtype=np.uint8)
        res_array1[10:20, 9:21, 9:21] = 1
        new_mask = calculate_mask(prop1, mask_base_array, None, (3, 1, 1))
        assert np.all(new_mask[0] == res_array1)
        res_array2 = np.zeros((30, 30, 30), dtype=np.uint8)
        res_array2[10:20, 8:22, 8:22] = 1
        res_array2[(9, 20), 9:21, 9:21] = 1
        res_array2[:, (8, 21, 8, 21), (8, 8, 21, 21)] = 0
        new_mask = calculate_mask(prop2, mask_base_array, None, (3, 1, 1))
        assert np.all(new_mask[0] == res_array2)
        res_array3 = np.zeros((30, 30, 30), dtype=np.uint8)
        res_array3[(9, 20), 8:22, 9:21] = 1
        res_array3[(9, 9, 20, 20), 9:21, (8, 21, 8, 21)] = 1

        res_array3[10:20, 7:23, 9:21] = 1
        res_array3[10:20, 8:22, (8, 21)] = 1
        res_array3[10:20, 9:21, (7, 22)] = 1
        new_mask = calculate_mask(prop3, mask_base_array, None, (3, 1, 1))
        assert np.all(new_mask == res_array3)

    def test_clip_mask(self):
        mask_base_array = np.zeros((30, 30, 30), dtype=np.uint8)
        mask_base_array[10:20, 10:20, 10:20] = 1
        mask2_array = np.copy(mask_base_array)
        mask2_array[13:17, 13:17, 13:17] = 0
        prop1 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=-0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )
        prop2 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=-0,
            fill_holes=RadiusType.NO,
            max_holes_size=70,
            save_components=False,
            clip_to_mask=True,
        )
        new_mask1 = calculate_mask(prop1, mask_base_array, mask2_array, (1, 1, 1))
        new_mask2 = calculate_mask(prop2, mask_base_array, mask2_array, (1, 1, 1))
        assert np.all(new_mask1 == mask_base_array)
        assert np.all(new_mask2 == mask2_array)

    def test_reversed_mask(self):
        mask_base_array = np.zeros((30, 30, 30), dtype=np.uint8)
        mask_base_array[10:20, 10:20, 10:20] = 1
        mask2_array = np.copy(mask_base_array)
        mask2_array[13:17, 13:17, 13:17] = 0
        prop1 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=-0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
            reversed_mask=False,
        )
        prop2 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=-0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
            reversed_mask=True,
        )
        new_mask1 = calculate_mask(prop1, mask_base_array, mask2_array, (1, 1, 1))
        new_mask2 = calculate_mask(prop2, mask_base_array, mask2_array, (1, 1, 1))
        assert np.all(new_mask1 == mask_base_array)
        assert np.all(new_mask2 == (mask_base_array == 0))


# TODO add Border rim and multiple otsu tests


class TestPipeline:
    @staticmethod
    def get_image():
        data = np.zeros((1, 50, 100, 100, 2), dtype=np.uint16)
        data[0, 10:40, 20:80, 20:60, 0] = 10
        data[0, 10:40, 20:80, 40:80, 1] = 10
        return Image(data, spacing=(100, 50, 50), file_path="", axes_order="TZYXC")

    @pytest.mark.parametrize("use_mask", [True, False])
    def test_pipeline_simple(self, use_mask):
        image = self.get_image()
        prop1 = MaskProperty(
            dilate=RadiusType.NO,
            dilate_radius=-0,
            fill_holes=RadiusType.NO,
            max_holes_size=0,
            save_components=False,
            clip_to_mask=False,
        )
        parameters1 = {
            "channel": 0,
            "minimum_size": 30,
            "threshold": {"name": "Manual", "values": {"threshold": 5}},
            "noise_filtering": {"name": "None", "values": {}},
            "side_connection": False,
        }
        parameters2 = {
            "channel": 1,
            "minimum_size": 30,
            "threshold": {"name": "Manual", "values": {"threshold": 5}},
            "noise_filtering": {"name": "None", "values": {}},
            "side_connection": False,
        }
        seg_profile1 = ROIExtractionProfile(name="Unknown", algorithm="Lower threshold", values=parameters1)
        pipeline_element = SegmentationPipelineElement(mask_property=prop1, segmentation=seg_profile1)
        seg_profile2 = ROIExtractionProfile(name="Unknown", algorithm="Lower threshold", values=parameters2)

        pipeline = SegmentationPipeline(name="test", segmentation=seg_profile2, mask_history=[pipeline_element])
        mask = np.ones(image.get_channel(0).shape, dtype=np.uint8) if use_mask else None
        result = calculate_pipeline(image=image, mask=mask, pipeline=pipeline, report_fun=empty)
        result_segmentation = np.zeros((50, 100, 100), dtype=np.uint8)
        result_segmentation[10:40, 20:80, 40:60] = 1
        assert np.all(result.roi_info.roi == result_segmentation)


class TestNoiseFiltering:
    @pytest.mark.parametrize("algorithm_name", NoiseFilterSelection.__register__.keys())
    def test_base(self, algorithm_name):
        noise_remove_algorithm = NoiseFilterSelection[algorithm_name]
        data = get_two_parts_array()[0, ..., 0]
        noise_remove_algorithm.noise_filter(data, (1, 1, 1), noise_remove_algorithm.get_default_values())


class TestConvexFill:
    def test_simple(self):
        arr = np.zeros((30, 30, 30), dtype=np.uint8)
        arr[10:-10, 10:-10, 10:-10] = 1
        res = convex_fill(arr)
        assert np.all(res == arr)
        arr = np.zeros((30, 60, 30), dtype=np.uint8)
        arr[10:-10, 10:-40, 10:-10] = 1
        arr[10:-10, -30:-10, 10:-10] = 2
        res = convex_fill(arr)
        assert np.all(res == arr)

    def test_missing_value(self):
        arr = np.zeros((30, 30, 30), dtype=np.uint8)
        arr[10:-10, 10:-10, 10:-10] = 2
        res = convex_fill(arr)
        assert np.all(res == arr)
        arr = np.zeros((30, 60, 30), dtype=np.uint8)
        arr[10:-10, 10:-40, 10:-10] = 1
        arr[10:-10, -30:-10, 10:-10] = 3
        res = convex_fill(arr)
        assert np.all(res == arr)

    def test_fill(self):
        arr = np.zeros((30, 30, 30), dtype=np.uint8)
        arr[10:-10, 10:-10, 10:-10] = 1
        arr2 = np.copy(arr)
        arr2[15:-15, 15:-15, 15:-15] = 0
        res = convex_fill(arr2)
        assert np.all(res == arr)
        arr2[15:-15, 15:-10, 15:-15] = 0
        res = convex_fill(arr2)
        assert np.all(res == arr)
        arr = np.zeros((30, 60, 30), dtype=np.uint8)
        arr[10:-10, 10:-40, 10:-10] = 1
        arr[10:-10, -30:-10, 10:-10] = 2
        arr2 = np.copy(arr)
        arr2[15:-15, 15:-15, 20:-10] = 0
        arr2[15:-15, -30:-10, 20:-10] = 0
        res = convex_fill(arr2)
        assert np.all(res == arr)

    def test_removed_object(self):
        arr = np.zeros((30, 30, 30), dtype=np.uint8)
        arr[10:-10, 10:-10, 10:-10] = 1
        arr2 = np.copy(arr)
        arr2[15:-15, 15:-15, 15:-15] = 2
        res = convex_fill(arr2)
        assert np.all(res == arr)
        arr2[15:-15, 15:-10, 15:-15] = 2
        res = convex_fill(arr2)
        assert np.all(res == arr)

    def test__convex_fill(self):
        arr = np.zeros((20, 20), dtype=bool)
        assert _convex_fill(arr) is None


class TestROIInfo:
    def test_none(self):
        si = ROIInfo(None)
        assert si.roi is None
        assert len(si.bound_info) == 0
        assert len(si.sizes) == 0

    def test_empty(self):
        si = ROIInfo(np.zeros((10, 10), dtype=np.uint8))
        assert np.all(si.roi == 0)
        assert len(si.bound_info) == 0
        assert len(si.sizes) == 1

    @pytest.mark.parametrize("num", [1, 5])
    def test_simple(self, num):
        data = np.zeros((10, 10), dtype=np.uint8)
        data[2:8, 2:8] = num
        si = ROIInfo(data)
        assert len(si.bound_info) == 1
        assert num in si.bound_info
        assert isinstance(si.bound_info[num], BoundInfo)
        assert np.all(si.bound_info[num].lower == [2, 2])
        assert np.all(si.bound_info[num].upper == [7, 7])
        assert len(si.sizes) == num + 1
        assert np.all(si.sizes[1:num] == 0)
        assert si.sizes[num] == 36

    @pytest.mark.parametrize("dims", [3, 5, 6])
    def test_more_dims(self, dims):
        si = ROIInfo(np.ones((10,) * dims, dtype=np.uint8))
        assert len(si.bound_info[1].lower) == dims
        assert len(si.bound_info[1].upper) == dims
        assert np.all(si.bound_info[1].lower == 0)
        assert np.all(si.bound_info[1].upper == 9)
        assert len(si.sizes) == 2
        assert np.all(si.sizes == [0, 10**dims])

    @pytest.mark.parametrize("comp_num", [2, 4, 8])
    def test_multiple_components(self, comp_num):
        data = np.zeros((10 * comp_num, 10), dtype=np.uint8)
        for i in range(comp_num):
            data[i * 10 + 2 : i * 10 + 8, 2:8] = i + 1
        si = ROIInfo(data)
        assert len(si.bound_info) == comp_num
        assert set(si.bound_info.keys()) == set(range(1, comp_num + 1))
        for i in range(comp_num):
            assert np.all(si.bound_info[i + 1].lower == [i * 10 + 2, 2])
            assert np.all(si.bound_info[i + 1].upper == [i * 10 + 7, 7])
        assert len(si.sizes) == comp_num + 1
        assert np.all(si.sizes[1:] == 36)

        data[-1, 8] = 1
        si = ROIInfo(data)
        assert np.all(si.bound_info[1].lower == 2)
        assert np.all(si.bound_info[1].upper == [10 * comp_num - 1, 8])

    def test_alternative(self):
        data = np.zeros((10, 10), dtype=np.uint8)
        data[2:8, 2:4] = 1
        data[2:8, 4:8] = 2
        alt1 = np.copy(data)
        alt1[data == 1] = 4
        alt1[data == 2] = 1
        alt2 = np.zeros((10, 10), dtype=np.uint8)

        ri = ROIInfo(data, alternative={"alt1": alt1, "alt2": alt2})
        assert ri.get_components_num("ROI") == 2
        assert ri.get_components_num("alt1") == 4
        assert ri.get_components_num("alt2") == 0
        assert ri.get_components_num("alt3") == 2


def test_bound_info():
    bi = BoundInfo(lower=np.array([1, 1, 1]), upper=np.array([5, 6, 7]))
    assert np.all(bi.box_size() == [5, 6, 7])
    assert len(bi.box_size()) == 3
    assert len(bi.get_slices()) == 3
    assert np.all([x == slice(1, 6 + i) for i, x in enumerate(bi.get_slices())])
    bi2 = bi.del_dim(1)
    assert len(bi2.box_size()) == 2
    assert len(bi2.get_slices()) == 2
    assert np.all(bi2.box_size() == [5, 7])


def test_dict_repr(monkeypatch):
    assert algorithm_base.dict_repr({}) == repr({})
    assert algorithm_base.dict_repr({1: 1}) == repr({1: 1})
    assert algorithm_base.dict_repr({1: 1, 2: {2: 2}}) == repr({1: 1, 2: {2: 2}})

    count = [0]

    def _repr(x):
        count[0] += 1
        return ""

    monkeypatch.setattr(algorithm_base, "numpy_repr", _repr)
    algorithm_base.dict_repr({1: np.zeros(5)})
    assert count[0] == 1
    count = [0]
    algorithm_base.dict_repr({1: np.zeros(5), 2: {1: np.zeros(5)}})
    assert count[0] == 2


def test_to_double_threshold():
    data = {
        "threshold": sa.TwoThreshold(
            lower_threshold=50,
            upper_threshold=100,
        )
    }
    data = sa._to_double_threshold(data)
    assert isinstance(data["threshold"], sa.DoubleThresholdSelection)



import os
import re
from functools import partial
from pathlib import Path
from queue import Queue
from typing import TYPE_CHECKING, NamedTuple, Optional, Union, cast

from pydantic import BaseModel
from qtpy.QtCore import QThread, Signal

from PartSeg._roi_mask.stack_settings import StackSettings, get_mask
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.mask.algorithm_description import MaskAlgorithmSelection
from PartSegCore.mask.io_functions import LoadROIImage, LoadStackImage, MaskProjectTuple, SaveROI
from PartSegCore.segmentation.algorithm_base import ROIExtractionAlgorithm

if TYPE_CHECKING:
    from PartSegCore.segmentation import StackAlgorithm


class BatchTask(NamedTuple):
    data: Union[str, MaskProjectTuple]
    parameters: ROIExtractionProfile
    save_prefix: Optional[tuple[Union[str, Path], Union[dict, BaseModel]]]


class BatchProceed(QThread):
    error_signal = Signal(str)
    progress_signal = Signal(str, int, str, int)
    range_signal = Signal(int, int)
    execution_done = Signal()
    multiple_result = Signal(MaskProjectTuple)
    algorithm: ROIExtractionAlgorithm

    def __init__(self):
        super().__init__()
        self.queue = Queue()
        self.algorithm = Optional[None]
        self.parameters = None
        self.file_list = []
        self.index = 0
        self.result_dir = ""
        self.save_parameters = {}

    def add_task(self, task: Union[BatchTask, list[BatchTask]]):
        if isinstance(task, list):
            for el in task:
                self.queue.put(el)
        else:
            self.queue.put(task)

    def progress_info(self, name: str, text: str, num: int):
        self.progress_signal.emit(text, num, name, self.index)

    def run_calculation(self):  # noqa: PLR0912  # FIXME
        while not self.queue.empty():
            task: BatchTask = self.queue.get()
            if isinstance(task.data, str):
                file_path = task.data
                if os.path.splitext(task.data)[1] in LoadROIImage.get_extensions():
                    project_tuple = LoadROIImage.load([task.data])
                else:
                    project_tuple = LoadStackImage.load([task.data])
            elif isinstance(task.data, MaskProjectTuple):
                project_tuple: MaskProjectTuple = task.data
                file_path = project_tuple.image.file_path
            else:
                continue
            try:
                name = os.path.basename(file_path)
                blank = get_mask(project_tuple.roi_info.roi, project_tuple.mask, project_tuple.selected_components)
                algorithm = cast("StackAlgorithm", MaskAlgorithmSelection[task.parameters.algorithm]())
                algorithm.set_image(project_tuple.image)
                algorithm.set_mask(blank)
                algorithm.set_parameters(task.parameters.values)
                if isinstance(task.save_prefix, tuple):
                    self.range_signal.emit(0, algorithm.get_steps_num() + 1)
                else:
                    self.range_signal.emit(0, algorithm.get_steps_num())
                # noinspection PyTypeChecker
                segmentation = algorithm.calculation_run(partial(self.progress_info, name))
                state2 = StackSettings.transform_state(
                    project_tuple,
                    segmentation.roi_info,
                    (
                        dict.fromkeys(segmentation.roi_info.bound_info, segmentation.parameters)
                        if segmentation.roi_info is not None
                        else {}
                    ),
                    [],
                )
                if isinstance(task.save_prefix, tuple):
                    self.progress_info(name, "saving", algorithm.get_steps_num())
                    name = f"{os.path.splitext(os.path.basename(file_path))[0]}.seg"
                    re_end = re.compile(r"(.*_version)(\d+)\.seg$")
                    os.makedirs(task.save_prefix[0], exist_ok=True)
                    while os.path.exists(os.path.join(task.save_prefix[0], name)):
                        if match := re_end.match(name):
                            num = int(match[2]) + 1
                            name = match[1] + str(num) + ".seg"
                        else:
                            name = f"{os.path.splitext(os.path.basename(file_path))[0]}_version1.seg"
                    SaveROI.save(os.path.join(task.save_prefix[0], name), state2, parameters=task.save_prefix[1])
                else:
                    self.multiple_result.emit(state2)
            except Exception as e:  # pylint: disable=broad-except
                self.error_signal.emit(f"Exception occurred during proceed {file_path}. Exception info {e}")
            self.index += 1
        self.index = 0
        self.execution_done.emit()

    def run(self):
        self.index = 0
        self.run_calculation()



"""from napari._qt.dialog.modal"""

from qtpy.QtCore import QPoint, QRect, Qt
from qtpy.QtGui import QCursor, QGuiApplication, QKeyEvent
from qtpy.QtWidgets import QDialog, QFrame, QVBoxLayout


class QtPopup(QDialog):
    """A generic popup window.

    The seemingly extra frame here is to allow rounded corners on a truly
    transparent background.  New items should be added to QtPopup.frame

    +----------------------------------
    | Dialog
    |  +-------------------------------
    |  | QVBoxLayout
    |  |  +----------------------------
    |  |  | QFrame
    |  |  |  +-------------------------
    |  |  |  |
    |  |  |  |  (add a new layout here)

    Parameters
    ----------
    parent : qtpy.QtWidgets:QWidget
        Parent widget of the popup dialog box.

    Attributes
    ----------
    frame : qtpy.QtWidgets.QFrame
        Frame of the popup dialog box.
    layout : qtpy.QtWidgets.QVBoxLayout
        Layout of the popup dialog box.
    """

    def __init__(self, parent):
        super().__init__(parent)
        self.setObjectName("QtModalPopup")
        self.setModal(False)  # if False, then clicking anywhere else closes it
        self.setWindowFlags(Qt.WindowType.Popup | Qt.WindowType.FramelessWindowHint)
        self.setLayout(QVBoxLayout())

        self.frame = QFrame()
        self.frame.setObjectName("QtPopupFrame")
        self.layout().addWidget(self.frame)
        self.layout().setContentsMargins(0, 0, 0, 0)

    def show_above_mouse(self, *args):
        """Show popup dialog above the mouse cursor position."""
        pos = QCursor().pos()  # mouse position
        szhint = self.sizeHint()
        pos -= QPoint(szhint.width() // 2, szhint.height() + 14)
        self.move(pos)
        self.show()

    def show_right_of_mouse(self, *args):
        pos = QCursor().pos()  # mouse position
        szhint = self.sizeHint()
        pos -= QPoint(-14, szhint.height() // 4)
        self.move(pos)
        self.show()

    def move_to(self, position="top", *, win_ratio=0.9, min_length=0):
        """Move popup to a position relative to the QMainWindow.

        Parameters
        ----------
        position : {str, tuple}, optional
            position in the QMainWindow to show the pop, by default 'top'
            if str: must be one of {'top', 'bottom', 'left', 'right' }
            if tuple: must be length 4 with (left, top, width, height)
        win_ratio : float, optional
            Fraction of the width (for position = top/bottom) or height (for
            position = left/right) of the QMainWindow that the popup will
            occupy.  Only valid when isinstance(position, str).
            by default 0.9
        min_length : int, optional
            Minimum size of the long dimension (width for top/bottom or
            height fort left/right).

        Raises
        ------
        ValueError
            if position is a string and not one of
            {'top', 'bottom', 'left', 'right' }
        """
        if isinstance(position, str):
            window = self.parent().window() if self.parent() else None
            if not window:
                raise ValueError("Specifying position as a string is only possible if the popup has a parent")
            left = window.pos().x()
            top = window.pos().y()
            if position in ("top", "bottom"):
                width = int(window.width() * win_ratio)
                width = max(width, min_length)
                left += (window.width() - width) // 2
                height = self.sizeHint().height()
                top += 24 if position == "top" else (window.height() - height - 12)
            elif position in ("left", "right"):
                height = int(window.height() * win_ratio)
                height = max(height, min_length)
                # 22 is for the title bar
                top += 22 + (window.height() - height) // 2
                width = self.sizeHint().width()
                left += 12 if position == "left" else (window.width() - width - 12)
            else:
                raise ValueError('position must be one of ["top", "left", "bottom", "right"]')
        elif isinstance(position, (tuple, list)):
            left, top, width, height = position
        else:
            raise ValueError(f"Wrong type of position {position}")

        # necessary for transparent round corners
        self.resize(self.sizeHint())
        # make sure the popup is completely on the screen
        # In Qt ≥5.10 we can use screenAt to know which monitor the mouse is on

        screen_geometry: QRect = QGuiApplication.screenAt(QCursor.pos()).geometry()

        left = max(min(screen_geometry.right() - width, left), screen_geometry.left())
        top = max(min(screen_geometry.bottom() - height, top), screen_geometry.top())
        self.setGeometry(left, top, width, height)

    def keyPressEvent(self, event: QKeyEvent):
        """Close window on return, else pass event through to super class.

        Parameters
        ----------
        event : qtpy.QtGui.QKeyEvent
            Event from the Qt context.
        """
        if event.key() in (Qt.Key.Key_Return, Qt.Key.Key_Enter):
            self.close()
            return
        super().keyPressEvent(event)



import contextlib
import json
import logging
import os
import tarfile
import typing
from contextlib import suppress
from copy import copy
from functools import partial
from io import BufferedIOBase, BytesIO, IOBase, RawIOBase, TextIOBase
from pathlib import Path
from threading import Lock

import numpy as np
import tifffile
from packaging.version import Version
from packaging.version import parse as parse_version
from tifffile import TiffFile

from PartSegCore.algorithm_describe_base import Register, ROIExtractionProfile
from PartSegCore.analysis import AnalysisAlgorithmSelection
from PartSegCore.analysis.io_utils import MaskInfo, ProjectTuple, project_version_info
from PartSegCore.io_utils import (
    IO_MASK_METADATA_FILE,
    LoadBase,
    LoadPoints,
    SegmentationType,
    WrongFileTypeException,
    check_segmentation_type,
    load_metadata_base,
    load_metadata_part,
    open_tar_file,
    proxy_callback,
    tar_to_buff,
)
from PartSegCore.json_hooks import partseg_object_hook
from PartSegCore.mask.io_functions import LoadROIImage
from PartSegCore.project_info import HistoryElement
from PartSegCore.roi_info import ROIInfo
from PartSegCore.universal_const import UNIT_SCALE, Units
from PartSegImage import GenericImageReader

__all__ = [
    "LoadImageForBatch",
    "LoadImageMask",
    "LoadMask",
    "LoadMaskSegmentation",
    "LoadProfileFromJSON",
    "LoadProject",
    "LoadStackImage",
    "load_dict",
    "load_metadata",
]

from PartSegImage.image import Image


def _load_history(tar_file):
    history = []
    with suppress(KeyError):
        history_buff = tar_file.extractfile(tar_file.getmember("history/history.json")).read()
        history_json = load_metadata(history_buff)
        for el in history_json:
            history_buffer = BytesIO()
            history_buffer.write(tar_file.extractfile(f"history/arrays_{el['index']}.npz").read())
            history_buffer.seek(0)
            el_up = update_algorithm_dict(el)
            segmentation_parameters = {"algorithm_name": el_up["algorithm_name"], "values": el_up["values"]}
            history.append(
                HistoryElement(
                    roi_extraction_parameters=segmentation_parameters,
                    mask_property=el_up["mask_property"],
                    arrays=history_buffer,
                    annotations=el_up.get("annotations", {}),
                )
            )
    return history


def load_project_from_tar(tar_file, file_path):
    if check_segmentation_type(tar_file) != SegmentationType.analysis:
        raise WrongFileTypeException
    image_buffer = BytesIO()
    image_tar = tar_file.extractfile(tar_file.getmember("image.tif"))
    image_buffer.write(image_tar.read())
    image_buffer.seek(0)
    reader = GenericImageReader()
    image = reader.read(image_buffer, ext=".tif")
    image.file_path = file_path

    algorithm_str = tar_file.extractfile("algorithm.json").read()
    algorithm_dict = load_metadata(algorithm_str)
    algorithm_dict = update_algorithm_dict(algorithm_dict)
    with contextlib.suppress(KeyError):
        algorithm_dict["algorithm_name"] = AnalysisAlgorithmSelection[algorithm_dict["algorithm_name"]].get_name()

    metadata = json.loads(tar_file.extractfile(IO_MASK_METADATA_FILE).read(), object_hook=partseg_object_hook)

    version = parse_version(metadata.get("project_version_info", "1.0"))

    if version == Version("1.0"):
        seg_dict = np.load(tar_to_buff(tar_file, "segmentation.npz"))
        mask = seg_dict.get("mask")
        roi = seg_dict["segmentation"]
    else:
        roi = tifffile.imread(tar_to_buff(tar_file, "segmentation.tif"))
        if "mask.tif" in tar_file.getnames():
            mask = tifffile.imread(tar_to_buff(tar_file, "mask.tif"))
            if np.max(mask) == 1:
                mask = mask.astype(bool)
        else:
            mask = None
    if "alternative.npz" in tar_file.getnames():
        alternative = np.load(tar_to_buff(tar_file, "alternative.npz"))
    else:
        alternative = {}
    history = _load_history(tar_file)
    image.set_mask(mask)
    roi_info = ROIInfo(roi, annotations=metadata.get("roi_annotations"), alternative=alternative)
    if version <= project_version_info:
        return ProjectTuple(
            file_path=file_path,
            image=image,
            roi_info=roi_info,
            mask=mask,
            history=history,
            algorithm_parameters=algorithm_dict,
        )
    logging.warning(
        "This project %(proj_ver) is from new version of PartSeg: %(version)",
        extra={"version": version, "proj_ver": project_version_info},
    )
    return ProjectTuple(
        file_path=file_path,
        image=image,
        roi_info=roi_info,
        mask=mask,
        history=history,
        algorithm_parameters=algorithm_dict,
        errors="This project is from new version of PartSeg. It may load incorrect.",
    )


def load_project(
    file: typing.Union[str, Path, tarfile.TarFile, TextIOBase, BufferedIOBase, RawIOBase, IOBase],
) -> ProjectTuple:
    """Load project from archive"""
    tar_file, file_path = open_tar_file(file)
    try:
        return load_project_from_tar(tar_file, file_path)
    finally:
        if isinstance(file, (str, Path)):
            tar_file.close()


class LoadProject(LoadBase):
    @classmethod
    def get_name(cls):
        return "Project (*.tgz *.tbz2 *.gz *.bz2)"

    @classmethod
    def get_short_name(cls):
        return "project"

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> ProjectTuple:
        return load_project(load_locations[0])


class LoadStackImage(LoadBase):
    @classmethod
    def get_name(cls):
        return "Image (*.tif *.tiff *.lsm *.czi *.oib *.oif *.obsep)"

    @classmethod
    def get_short_name(cls):
        return "tiff_image"

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ):
        if metadata is None:
            metadata = {"default_spacing": tuple(1 / UNIT_SCALE[Units.nm.value] for _ in range(3))}

        if "recursion_limit" not in metadata:
            metadata = copy(metadata)
            metadata["recursion_limit"] = 3
        image = GenericImageReader.read_image(
            load_locations[0],
            callback_function=partial(proxy_callback, range_changed, step_changed),
            default_spacing=tuple(metadata["default_spacing"]),
        )
        re_read = all(el[0] == el[1] for el in image.get_ranges())
        if re_read and metadata["recursion_limit"] > 0:
            metadata["recursion_limit"] -= 1
            cls.load(load_locations, range_changed, step_changed, metadata)
        return ProjectTuple(load_locations[0], image)


class LoadImageMask(LoadBase):
    @classmethod
    def get_name(cls):
        return "Image with mask (*.tif *.tiff *.lsm)"

    @classmethod
    def get_short_name(cls):
        return "image_with_mask"

    @classmethod
    def number_of_files(cls):
        return 2

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ):
        if metadata is None:
            metadata = {"default_spacing": (10**-6, 10**-6, 10**-6)}
        if len(load_locations) == 1:
            new_path, ext = os.path.splitext(load_locations[0])
            new_path += f"_mask{ext}"
            if not os.path.exists(new_path):
                raise ValueError("Cannot determine mask file. It need to have '_mask' suffix.")
            load_locations.append(load_locations)
        image = GenericImageReader.read_image(
            load_locations[0],
            load_locations[1],
            callback_function=partial(proxy_callback, range_changed, step_changed),
            default_spacing=tuple(metadata["default_spacing"]),
        )
        return ProjectTuple(load_locations[0], image, mask=image.mask)

    @classmethod
    def get_next_file(cls, file_paths: list[str]):
        base, ext = os.path.splitext(file_paths[0])
        return f"{base}_mask{ext}"


class LoadMask(LoadBase):
    @classmethod
    def get_name(cls):
        return "mask to image (*.tif *.tiff)"

    @classmethod
    def get_short_name(cls):
        return "mask_to_name"

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ):
        image_file = TiffFile(load_locations[0])
        count_pages = [0]
        mutex = Lock()

        def report_func():
            mutex.acquire()
            count_pages[0] += 1
            step_changed(count_pages[0])
            mutex.release()

        range_changed(0, len(image_file.series[0]))
        image_file.report_func = report_func
        mask_data = image_file.asarray()
        return MaskInfo(load_locations[0], mask_data)

    @classmethod
    def partial(cls):
        return True


def _mask_data_outside_mask(file_path):
    if not isinstance(file_path, str):
        return False
    with tarfile.open(file_path, "r:*") as tar_file:
        metadata = load_metadata_base(tar_file.extractfile(IO_MASK_METADATA_FILE).read().decode("utf8"))
        return metadata.get("keep_data_outside_mask", False)


def load_mask_project(
    load_locations: list[typing.Union[str, BytesIO, Path]],
    range_changed: typing.Callable[[int, int], typing.Any],
    step_changed: typing.Callable[[int], typing.Any],
    metadata: typing.Optional[dict] = None,
):
    data = LoadROIImage.load(load_locations, range_changed, step_changed, metadata)
    zero_out_cut_area = _mask_data_outside_mask(load_locations[0])
    image = data.image
    if not isinstance(image, Image):  # pragma: no cover
        raise ValueError("Image is not instance of Image class.")
    if data.roi_info.roi is None:  # pragma: no cover
        raise ValueError("No ROI found in the image.")
    roi = data.roi_info.roi
    components = data.selected_components
    if not components:
        components = list(data.roi_info.bound_info)
    res = []
    base, ext = os.path.splitext(load_locations[0])
    range_changed(0, len(components))
    str_len = str(len(str(len(components))))
    path_template = base + "_component{:0" + str_len + "d}" + ext
    for i in components:
        step_changed(i)
        bound = data.roi_info.bound_info[i]
        single_roi = roi[tuple(bound.get_slices(data.frame_thickness))] == i
        if not np.any(single_roi):
            continue
        im = image.cut_image(
            bound.get_slices(), replace_mask=True, zero_out_cut_area=zero_out_cut_area, frame=data.frame_thickness
        ).cut_image(
            single_roi,
            replace_mask=True,
            zero_out_cut_area=zero_out_cut_area,
            frame=data.frame_thickness,
        )
        im.file_path = path_template.format(i)
        res.append(ProjectTuple(im.file_path, im, mask=im.mask))
    return res


class LoadMaskSegmentation(LoadBase):
    @classmethod
    def get_name(cls):
        return "Mask project (*.seg *.tgz)"

    @classmethod
    def get_short_name(cls):
        return "mask_project"

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> list[ProjectTuple]:
        if range_changed is None:

            def range_changed(_x, _y):
                return None

        if step_changed is None:

            def step_changed(_):
                return None

        return load_mask_project(load_locations, range_changed, step_changed, metadata)


class LoadProfileFromJSON(LoadBase):
    @classmethod
    def get_short_name(cls):
        return "json"

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> tuple[dict, list]:
        return load_metadata_part(load_locations[0])

    @classmethod
    def get_name(cls) -> str:
        return "Segment profile (*.json)"


def load_metadata(data: typing.Union[str, Path]):
    """
    Load metadata saved in json format for segmentation mask

    :param data: path to json file, string with json, or opened file
    :return: restored structures
    """
    return load_metadata_base(data)


def update_algorithm_dict(dkt):
    if "name" in dkt:
        profile = ROIExtractionProfile(name="", algorithm=dkt["name"], values=dkt["values"])
    elif "algorithm_name" in dkt:
        profile = ROIExtractionProfile(name="", algorithm=dkt["algorithm_name"], values=dkt["values"])
    else:
        return dkt
    res = dict(dkt)
    res.update({"algorithm_name": profile.algorithm, "values": profile.values})
    return res


load_dict = Register(
    LoadStackImage,
    LoadImageMask,
    LoadProject,
    LoadMaskSegmentation,
    LoadPoints,
    class_methods=LoadBase.need_functions,
)


class LoadImageForBatch(LoadBase):
    @classmethod
    def get_short_name(cls):
        return "load_all"

    @classmethod
    def load(
        cls,
        load_locations: list[typing.Union[str, BytesIO, Path]],
        range_changed: typing.Optional[typing.Callable[[int, int], typing.Any]] = None,
        step_changed: typing.Optional[typing.Callable[[int], typing.Any]] = None,
        metadata: typing.Optional[dict] = None,
    ) -> typing.Union[ProjectTuple, list[ProjectTuple]]:
        ext = os.path.splitext(load_locations[0])[1].lower()

        for loader in load_dict.values():
            if loader.partial() or loader.number_of_files() != 1:
                continue
            if ext in loader.get_extensions():
                res = loader.load([load_locations[0]], metadata=metadata)
                if isinstance(res, list):
                    return [cls._clean_project(x) for x in res]
                return cls._clean_project(res)
        raise ValueError(f"Cannot load file {load_locations[0]}")

    @staticmethod
    def _clean_project(project: ProjectTuple):
        return ProjectTuple(file_path=project.file_path, image=project.image.substitute(mask=None))

    @classmethod
    def get_name(cls) -> str:
        ext_set = set()
        for loader in load_dict.values():
            if loader.partial() or loader.number_of_files() != 1:
                continue
            ext_set.update(loader.get_extensions())

        return f"Load generic ({' '.join(f'*{x}' for x in ext_set)})"



import json
import urllib.error
import urllib.request
from datetime import date, timedelta
from io import StringIO
from unittest.mock import MagicMock

import packaging.version
import pytest
from qtpy.QtWidgets import QMessageBox

from PartSeg import state_store
from PartSeg._launcher import check_version
from PartSeg._launcher.check_version import IGNORE_FILE


@pytest.mark.enablethread
@pytest.mark.parametrize("thread", [False, True])
@pytest.mark.parametrize("package_name", ["PartSeg", "sample_name"])
def test_fetching(thread, package_name, monkeypatch, qtbot):
    def urlopen_mock(url):
        assert f"https://pypi.org/pypi/{package_name}/json" == url
        return StringIO(
            json.dumps(
                {"info": {"version": "0.10.0", "home_page": f"https://4dnucleome.cent.uw.edu.pl/{package_name}/"}}
            )
        )

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    assert packaging.version.parse("0.10.0") < packaging.version.parse("0.11.0")
    chk_thr = check_version.CheckVersionThread(package_name, base_version="0.11.0")
    chk_thr.release = "0.10.0"
    if thread:
        with qtbot.wait_signal(chk_thr.finished):
            chk_thr.start()
    else:
        chk_thr.run()
    assert chk_thr.release == "0.10.0"
    assert chk_thr.url == f"https://4dnucleome.cent.uw.edu.pl/{package_name}/"
    chk_thr.deleteLater()


@pytest.mark.parametrize("frozen", [True, False])
def test_show_window_dialog(monkeypatch, frozen, qtbot):
    values = ["", ""]

    class MockMessageBox:
        StandardButton = QMessageBox.StandardButton

        Icon = QMessageBox.Icon

        def __init__(self, _type, title, message, _buttons):
            values[0] = title
            values[1] = message

        @staticmethod
        def exec_():
            return check_version.QMessageBox.StandardButton.Ok

    chk_thr = check_version.CheckVersionThread(base_version="0.10.0")
    chk_thr.release = "0.11.0"
    monkeypatch.setattr(check_version.sys, "frozen", frozen, raising=False)
    monkeypatch.setattr(check_version, "QMessageBox", MockMessageBox)
    chk_thr.show_version_info()
    assert values[0] == "New release"
    if frozen:
        assert "You can download next release form" in values[1]
    else:
        assert "You can update it from pypi" in values[1]


def test_no_update(monkeypatch, qtbot, tmp_path):
    # check if nothing is reported on ulr error
    monkeypatch.setattr("PartSeg.state_store.save_folder", tmp_path)

    def urlopen_mock(url):
        raise urllib.error.URLError("test purpose")

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    chk_thr = check_version.CheckVersionThread(base_version="0.10.0")
    chk_thr.run()
    assert chk_thr.release == "0.10.0"


def test_error_report(monkeypatch, qtbot):
    sentry_val = [False]

    def urlopen_mock(url):
        raise RuntimeError

    def sentry_mock(e):
        sentry_val[0] = True

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)
    monkeypatch.setattr(check_version.sentry_sdk, "capture_exception", sentry_mock)
    chk_thr = check_version.CheckVersionThread()
    chk_thr.run()

    assert sentry_val[0] is True


def test_ignore_file_exists(monkeypatch, qtbot, tmp_path):
    with (tmp_path / IGNORE_FILE).open("w") as f_p:
        f_p.write(date.today().isoformat())

    monkeypatch.setattr(state_store, "save_folder", tmp_path)

    def urlopen_mock(url):
        raise RuntimeError

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)

    chk_thr = check_version.CheckVersionThread()
    chk_thr.run()


def test_ignore_file_exists_old_date(monkeypatch, qtbot, tmp_path):
    with (tmp_path / IGNORE_FILE).open("w") as f_p:
        f_p.write((date.today() - timedelta(days=60)).isoformat())

    monkeypatch.setattr(state_store, "save_folder", tmp_path)

    def urlopen_mock(_url):
        return StringIO('{"info": {"version": "0.11.0", "home_page": "example.org"}}')

    monkeypatch.setattr(urllib.request, "urlopen", urlopen_mock)

    chk_thr = check_version.CheckVersionThread(base_version="0.10.0")
    chk_thr.run()
    assert not (tmp_path / IGNORE_FILE).exists()
    assert chk_thr.release == "0.11.0"
    assert chk_thr.url == "example.org"


def test_create_ignore(qtbot, tmp_path, monkeypatch):
    monkeypatch.setattr(state_store, "save_folder", tmp_path)

    chk_thr = check_version.CheckVersionThread(base_version="0.10.0")
    chk_thr.release = "0.11.0"
    monkeypatch.setattr(check_version.QMessageBox, "exec_", MagicMock(return_value=check_version.QMessageBox.Ignore))
    chk_thr.show_version_info()
    assert (tmp_path / IGNORE_FILE).exists()
    with (tmp_path / IGNORE_FILE).open() as f_p:
        assert f_p.read() == date.today().isoformat()



# pylint: disable=no-self-use
import math
import sys
import typing
from enum import Enum

import pytest
from local_migrator import class_to_str, register_class
from pydantic import BaseModel as PydanticBaseModel
from pydantic import Field, ValidationError

from PartSegCore.algorithm_describe_base import (
    AlgorithmDescribeBase,
    AlgorithmProperty,
    AlgorithmSelection,
    ROIExtractionProfile,
    _GetDescriptionClass,
    base_model_to_algorithm_property,
)
from PartSegCore.analysis import AnalysisAlgorithmSelection
from PartSegCore.segmentation.restartable_segmentation_algorithms import LowerThresholdAlgorithm
from PartSegCore.utils import BaseModel
from PartSegImage import Channel


def test_algorithm_property():
    ap = AlgorithmProperty("test", "Test", 1)
    assert ap.name == "test"
    assert "user_name='Test'" in repr(ap)


def test_algorithm_property_warn():
    with pytest.warns(DeprecationWarning, match="use value_type instead"):
        ap = AlgorithmProperty("test", "Test", 1, property_type=int)
    assert ap.value_type is int


def test_algorithm_property_no_kwargs():
    with pytest.raises(ValueError, match="are not expected"):
        AlgorithmProperty("test", "Test", 1, a=1)


def test_algorithm_property_list_exc():
    with pytest.raises(ValueError, match="should be one of possible values"):
        AlgorithmProperty("test", "Test", 1, possible_values=[2, 3], value_type=list)


def test_get_description_class():
    class SampleClass:
        __test_class__ = _GetDescriptionClass()

        @classmethod
        def get_fields(cls):
            return [AlgorithmProperty("test1", "Test 1", 1), AlgorithmProperty("test2", "Test 2", 2.0)]

    val = SampleClass.__test_class__
    assert val.__name__ == "__test_class__"
    assert val.__qualname__.endswith("SampleClass.__test_class__")
    assert issubclass(val, PydanticBaseModel)
    assert val.__fields__.keys() == {"test1", "test2"}


def test_algorithm_selection():
    class TestSelection(AlgorithmSelection):
        pass

    class TestSelection2(AlgorithmSelection):
        pass

    class Class1(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls) -> str:
            return "test1"

        @classmethod
        def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
            return []

    class Class2(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls) -> str:
            return "test2"

        @classmethod
        def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
            return []

    TestSelection.register(Class1)
    TestSelection.register(Class2)

    assert "test1" in TestSelection.__register__
    assert "test1" not in TestSelection2.__register__

    v = TestSelection(name="test1", values={})
    assert v.name == "test1"
    assert v.class_path == class_to_str(Class1)
    assert v.values == {}

    with pytest.raises(ValidationError):
        TestSelection(name="test3", values={})

    assert TestSelection["test1"] is Class1

    assert TestSelection.__register__ != TestSelection2.__register__

    ts = TestSelection(name="test1", values={})
    assert ts.algorithm() == Class1


def test_register_errors():
    class TestSelection(AlgorithmSelection):
        pass

    class Alg1:
        pass

    class Alg2(AlgorithmDescribeBase):
        pass

    class Alg3(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls):
            return 1

        @classmethod
        def get_fields(cls):
            return []  # pragma: no cover

    with pytest.raises(ValueError, match="Class .* need to define classmethod 'get_name'"):
        TestSelection.register(Alg1)

    with pytest.raises(ValueError, match="Class .* need to implement classmethod 'get_name'"):
        TestSelection.register(Alg2)

    with pytest.raises(ValueError, match="Class .* need to implement classmethod 'get_name'"):
        TestSelection.__register__["test1"] = Alg2

    with pytest.raises(ValueError, match="Function get_name of class .* need return string not .*int"):
        TestSelection.register(Alg3)


def test_register_name_collision():
    class TestSelection(AlgorithmSelection):
        pass

    class Alg1(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls):
            return "1"

        @classmethod
        def get_fields(cls):
            return []

    class Alg2(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls):
            return "1"

        @classmethod
        def get_fields(cls):
            return []  # pragma: no cover

    class Alg3(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls):
            return "2"

        @classmethod
        def get_fields(cls):
            return []

    TestSelection.register(Alg1, old_names=["0"])
    with pytest.raises(
        ValueError, match="Object .* with this name: '1' already exist and register is not in replace mode"
    ):
        TestSelection.register(Alg2)

    assert len(TestSelection.__register__) == 1

    with pytest.raises(ValueError, match="Old value mapping for name '0' already registered"):
        TestSelection.register(Alg3, old_names=["0"])


def test_register_not_subclass():
    class TestSelection(AlgorithmSelection):
        pass

    class Alg1:
        @classmethod
        def get_name(cls):
            return "1"

        @classmethod
        def get_fields(cls):
            return []  # pragma: no cover

    with pytest.raises(ValueError, match="Class .* need to be subclass of .*AlgorithmDescribeBase"):
        TestSelection.register(Alg1)


def test_register_validate_name_assignment():
    class TestSelection(AlgorithmSelection):
        pass

    class Alg1(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls):
            return "1"

        @classmethod
        def get_fields(cls):
            return []  # pragma: no cover

    class Alg2(Alg1):
        @classmethod
        def get_name(cls):
            return 2

    with pytest.raises(ValueError, match="need return string"):
        TestSelection.__register__["1"] = Alg2

    with pytest.raises(ValueError, match="under name returned by get_name function"):
        TestSelection.__register__["2"] = Alg1


def test_register_get_fields_validity():
    class TestSelection(AlgorithmSelection):
        pass

    class Alg1(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls):
            return "1"

        @classmethod
        def get_fields(cls):
            raise NotImplementedError

    class Alg2(Alg1):
        @classmethod
        def get_fields(cls):
            return ()

    with pytest.raises(ValueError, match="need to be implemented"):
        TestSelection.register(Alg1)
    with pytest.raises(ValueError, match="need return list not"):
        TestSelection.register(Alg2)


def test_register_no_default_value():
    class TestSelection(AlgorithmSelection):
        pass

    with pytest.raises(ValueError, match="Register does not contain any algorithm"):
        TestSelection.get_default()


def test_algorithm_selection_convert_subclass(clean_register):
    class TestSelection(AlgorithmSelection):
        pass

    @register_class
    class TestModel1(BaseModel):
        field1: int = 0

    @register_class(version="0.0.1", migrations=[("0.0.1", lambda x: {"field2": x["field"]})])
    class TestModel2(BaseModel):
        field2: int = 7

    class Class1(AlgorithmDescribeBase):
        __argument_class__ = TestModel1

        @classmethod
        def get_name(cls) -> str:
            return "test1"

    class Class2(AlgorithmDescribeBase):
        __argument_class__ = TestModel2

        @classmethod
        def get_name(cls) -> str:
            return "test2"

    TestSelection.register(Class1)
    TestSelection.register(Class2)

    ob = TestSelection(name="test1", values={"field1": 4})
    assert isinstance(ob.values, TestModel1)
    assert ob.values.field1 == 4

    ob = TestSelection(name="test2", values={"field": 5})
    assert isinstance(ob.values, TestModel2)
    assert ob.values.field2 == 5


def test_algorithm_selection_register_old(clean_register):
    class TestSelection(AlgorithmSelection):
        pass

    @register_class
    class TestModel1(BaseModel):
        field1: int = 0

    class Class1(AlgorithmDescribeBase):
        __argument_class__ = TestModel1

        @classmethod
        def get_name(cls) -> str:
            return "test3"

    TestSelection.register(Class1, old_names=["test"])

    ob = TestSelection(name="test3", values={"field1": 4})
    assert isinstance(ob.values, TestModel1)
    ob = TestSelection(name="test", values={"field1": 4})
    assert isinstance(ob.values, TestModel1)
    with pytest.raises(ValidationError):
        TestSelection(name="test4", values={"field1": 4})


def test_base_model_to_algorithm_property_base():
    class SampleEnum(Enum):
        a = 1
        b = 2

    class Sample(BaseModel):
        field1: int = Field(0, le=100, ge=0, title="Field 1")
        field2: SampleEnum = SampleEnum.a
        field_3: float = Field(0, le=55, ge=-7)
        channel: Channel = Field(0, title="Channel")

    s = Sample(field1=1, field_3=1.5)
    assert s.field_3 == 1.5

    converted = base_model_to_algorithm_property(Sample)
    assert len(converted) == 4
    assert converted[0].name == "field1"
    assert converted[0].user_name == "Field 1"
    assert issubclass(converted[0].value_type, int)
    assert converted[0].range == (0, 100)
    assert converted[1].name == "field2"
    assert converted[1].user_name == "Field2"
    assert converted[1].value_type is SampleEnum
    assert converted[2].name == "field_3"
    assert converted[2].user_name == "Field 3"
    assert issubclass(converted[2].value_type, float)
    assert converted[2].range == (-7, 55)

    assert converted[3].value_type is Channel
    assert converted[3].name == "channel"
    assert converted[3].user_name == "Channel"


@pytest.mark.skipif(sys.version_info < (3, 9), reason="requires python 3.9 or higher")
def test_gt_lt_conversion():
    class Sample(BaseModel):
        field1: int = Field(0, le=100, ge=0, title="Field 1")
        field2: int = Field(0, lt=100, gt=0, title="Field 2")
        field3: float = Field(0, le=100, ge=0, title="Field 3")
        field4: float = Field(0, lt=100, gt=0, title="Field 4")
        field5: int = Field(0, title="Field 5")

    converted = base_model_to_algorithm_property(Sample)
    assert converted[0].range == (0, 100)
    assert converted[1].range == (1, 99)
    assert converted[2].range == (0, 100)
    assert converted[3].range == (math.nextafter(0, math.inf), math.nextafter(100, -math.inf))
    assert converted[4].range == (0, 1000)


def test_base_model_to_algorithm_property_algorithm_describe_base():
    class SampleSelection(AlgorithmSelection):
        pass

    class SampleClass1(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls) -> str:
            return "1"

        @classmethod
        def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
            return []

    class SampleClass2(AlgorithmDescribeBase):
        @classmethod
        def get_name(cls) -> str:
            return "2"

        @classmethod
        def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
            return []

    SampleSelection.register(SampleClass1)
    SampleSelection.register(SampleClass2)

    d_text = "description text"

    class SampleModel(BaseModel):
        field1: int = Field(10, le=100, ge=0, title="Field 1", description=d_text)
        check_selection: SampleSelection = Field(SampleSelection(name="1", values={}), title="Class selection")

    converted = base_model_to_algorithm_property(SampleModel)
    assert len(converted) == 2
    assert issubclass(converted[0].value_type, int)
    assert converted[0].help_text == d_text
    assert issubclass(converted[1].value_type, AlgorithmDescribeBase)
    assert converted[1].default_value == "1"
    assert converted[1].possible_values is SampleSelection.__register__


def test_base_model_to_algorithm_property_algorithm_describe_empty():
    assert base_model_to_algorithm_property(BaseModel) == []


def test_text_addition_model_to_algorithm_property():
    class ModelWithText(BaseModel):
        field1: int = 1
        field2: int = Field(1, prefix="aaaa")
        field3: int = Field(1, suffix="bbbb")
        field4: int = 1

        @staticmethod
        def header():
            return "cccc"

    property_list = base_model_to_algorithm_property(ModelWithText)
    assert property_list[0] == "cccc"
    assert property_list[2] == "aaaa"
    assert property_list[5] == "bbbb"


def test_base_model_to_algorithm_property_position():
    class BBaseModel(BaseModel):
        field1: int = 1
        field2: int = 1

    class ModelWithPosition(BBaseModel):
        field3: int = Field(1, position=1)

    property_list = base_model_to_algorithm_property(ModelWithPosition)
    assert property_list[0].name == "field1"
    assert property_list[1].name == "field3"
    assert property_list[2].name == "field2"


def test_base_model_to_algorithm_property_magicgui_parameters():
    class BBaseModel(BaseModel):
        field1: int = Field(1, options={"a": 1, "b": 2})

    prop = base_model_to_algorithm_property(BBaseModel)[0]
    assert prop.mgi_options == {"a": 1, "b": 2}


def test_base_model_to_algorithm_property_hline():
    class Model(BaseModel):
        field1: int = 1
        field2: int = Field(1, prefix="------", suffix="---", position=0)

    fields = base_model_to_algorithm_property(Model)

    assert len(fields) == 4
    assert isinstance(fields[0], str)
    assert isinstance(fields[2], str)


def test_hidden_field():
    class Model(BaseModel):
        field1: int = Field(1, hidden=True)
        field2: int = 1

    fields = base_model_to_algorithm_property(Model)
    assert len(fields) == 1
    assert fields[0].name == "field2"


class TestAlgorithmDescribeBase:
    def test_old_style_algorithm(self):
        class SampleAlgorithm(AlgorithmDescribeBase):
            @classmethod
            def get_name(cls) -> str:
                return "sample"

            @classmethod
            def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
                return ["aaaa", AlgorithmProperty("name", "Name", 1, options_range=(1, 10), help_text="ceeeec")]

        assert SampleAlgorithm.get_name() == "sample"
        assert len(SampleAlgorithm.get_fields()) == 2
        assert "ceeeec" in SampleAlgorithm.get_doc_from_fields()
        assert "(default values: 1)" in SampleAlgorithm.get_doc_from_fields()
        assert len(SampleAlgorithm.get_fields_dict()) == 1
        assert SampleAlgorithm.get_default_values() == {"name": 1}

    def test_new_style_algorithm(self):
        class DataModel(BaseModel):
            name: int = Field(1, ge=1, le=10, description="ceeeec", prefix="aaaa")

        class SampleAlgorithm(AlgorithmDescribeBase):
            __argument_class__ = DataModel

            @classmethod
            def get_name(cls) -> str:
                return "sample"

        assert SampleAlgorithm.get_name() == "sample"
        with pytest.warns(FutureWarning, match=r"Class has __argument_class__ defined"):
            assert len(SampleAlgorithm.get_fields()) == 2
        assert "ceeeec" in SampleAlgorithm.get_doc_from_fields()
        assert "(default values: 1)" in SampleAlgorithm.get_doc_from_fields()
        assert len(SampleAlgorithm.get_fields_dict()) == 1
        assert SampleAlgorithm.get_default_values() == DataModel(name=1)

    def test_new_style_algorithm_with_old_style_subclass(self):
        class DataModel(BaseModel):
            name: int = Field(1, ge=1, le=10, description="ceeeec", prefix="aaaa")

        class SampleAlgorithm(AlgorithmDescribeBase):
            __argument_class__ = DataModel

            @classmethod
            def get_name(cls) -> str:
                return "sample"

        class SampleSubAlgorithm(SampleAlgorithm):
            @classmethod
            def get_name(cls) -> str:
                return "sample2"

            @classmethod
            def get_fields(cls) -> list[typing.Union[AlgorithmProperty, str]]:
                return [
                    *super().get_fields(),
                    AlgorithmProperty("name2", "Name 2", 3.0, options_range=(1, 10), help_text="deeeed"),
                ]

        assert SampleSubAlgorithm.get_name() == "sample2"
        assert SampleAlgorithm.get_name() == "sample"
        with pytest.warns(FutureWarning, match=r"Class has __argument_class__ defined"):
            assert len(SampleSubAlgorithm.get_fields()) == 3
        with pytest.warns(FutureWarning, match=r"Class has __argument_class__ defined"):
            doc_text = SampleSubAlgorithm.get_doc_from_fields()
        assert "ceeeec" in doc_text
        assert "deeeed" in doc_text
        assert "(default values: 1)" in doc_text
        assert "(default values: 3.0)" in doc_text
        with pytest.warns(FutureWarning, match=r"Class has __argument_class__ defined"):
            assert len(SampleSubAlgorithm.get_fields_dict()) == 2
        with pytest.warns(FutureWarning, match=r"Class has __argument_class__ defined"):
            assert SampleSubAlgorithm.get_default_values() == {"name": 1, "name2": 3.0}


class TestROIExtractionProfile:
    def test_roi_extraction_profile(self):
        ROIExtractionProfile(name="aaa", algorithm="aaa", values={})
        with pytest.warns(FutureWarning):
            ROIExtractionProfile("aaa", "aaa", {})  # pylint: disable=too-many-function-args

    def test_dump_dict(self):
        prof = ROIExtractionProfile(
            name="aaa",
            algorithm=LowerThresholdAlgorithm.get_name(),
            values=LowerThresholdAlgorithm.get_default_values(),
        )
        assert prof.values.threshold.values.threshold == 8000
        assert prof.dict()["values"]["threshold"]["values"]["threshold"] == 8000

    def test_pretty_print(self):

        prof1 = ROIExtractionProfile(name="aaa", algorithm="Lower threshold", values={})
        assert prof1.pretty_print(AnalysisAlgorithmSelection).startswith("ROI extraction profile name:")
        prof1 = ROIExtractionProfile(name="", algorithm="Lower threshold", values={})
        assert prof1.pretty_print(AnalysisAlgorithmSelection).startswith("ROI extraction profile\n")
        prof2 = ROIExtractionProfile(
            name="aaa",
            algorithm=LowerThresholdAlgorithm.get_name(),
            values=LowerThresholdAlgorithm.get_default_values(),
        )
        assert prof2.pretty_print(AnalysisAlgorithmSelection).count("\n") == 7



import os
from importlib import metadata

from packaging.version import parse as parse_version
from qtpy import QT_VERSION
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QGridLayout, QLabel, QPushButton

import PartSeg
from PartSeg.common_gui.universal_gui_part import TextShow


def get_links():
    source_code = "https://github.com/4DNucleome/PartSeg"
    homepage = "https://partseg.github.io/"
    for item in metadata.metadata("PartSeg").get_all("Project-URL"):
        if item.startswith("Homepage"):
            homepage = item.split(" ", 1)[1]
        elif item.startswith("Source"):
            source_code = item.rsplit(" ", 1)[1]

    return homepage, source_code


class AboutDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        homepage, source_code = get_links()
        self.setWindowTitle("About PartSeg")
        text = (
            f"<strong>PartSeg</strong> ({PartSeg.__version__})<br>"
            "PartSeg is gui and library for segmentation algorithms on high resolution microscopy<br><br>Webpage: "
            f"<a href='{homepage}'>{homepage}</a>"
            "<br>Repository and issue tracker: "
            f"<a href='{source_code}'>{source_code}</a>"
        )
        cite_as_text = (
            "Bokota, G., Sroka, J., Basu, S. et al. PartSeg: a tool for quantitative feature"
            " extraction from 3D microscopy images for dummies. BMC Bioinformatics 22, 72 (2021)."
            " https://doi.org/10.1186/s12859-021-03984-1"
        )
        dev_path = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(PartSeg.__file__))), "changelog.md")
        if os.path.exists(dev_path):
            with open(dev_path, encoding="utf-8") as ff:
                changelog_text = ff.read()
        else:
            changelog_text = PartSeg.changelog
        text_label = QLabel(self)
        text_label.setText(text)
        self.change_log = TextShow()
        self.change_log.setAcceptRichText(True)
        if parse_version(QT_VERSION) < parse_version("5.14.0"):  # pragma: no cover
            self.change_log.setText(changelog_text)
        else:
            self.change_log.setMarkdown(changelog_text)
        self.cite_as = TextShow(lines=3)
        if parse_version(QT_VERSION) < parse_version("5.14.0"):  # pragma: no cover
            self.cite_as.setText(cite_as_text)
        else:
            self.cite_as.setMarkdown(cite_as_text)
        self.cite_as.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.LinksAccessibleByMouse
        )
        ok_but = QPushButton("Ok")
        ok_but.clicked.connect(self.accept)
        text_label.setTextInteractionFlags(
            Qt.TextInteractionFlag.TextSelectableByMouse | Qt.TextInteractionFlag.LinksAccessibleByMouse
        )
        text_label.setOpenExternalLinks(True)
        layout = QGridLayout(self)
        layout.addWidget(text_label, 0, 0, 1, 3)
        layout.addWidget(self.change_log, 1, 0, 1, 3)
        layout.addWidget(QLabel("Cite as:"), 2, 0, 1, 3)
        layout.addWidget(self.cite_as, 3, 0, 1, 3)
        layout.addWidget(ok_but, 4, 2)
        self.setLayout(layout)



from vispy.app import MouseEvent

from PartSeg._roi_mask.stack_settings import StackSettings
from PartSeg.common_gui.channel_control import ChannelProperty
from PartSeg.common_gui.napari_image_view import ImageInfo, ImageView, LabelEnum


class StackImageView(ImageView):
    """
    :cvar settings: StackSettings
    """

    settings: StackSettings

    def __init__(self, settings: StackSettings, channel_property: ChannelProperty, name: str):
        super().__init__(settings, channel_property, name)
        self.viewer_widget.canvas.events.mouse_press.connect(self.component_click)

    def refresh_selected(self):
        if (
            self.settings.get_from_profile(f"{self.name}.image_state.show_label", LabelEnum.Show_results)
            == LabelEnum.Show_selected
        ):
            self.update_roi_labeling()

    def component_click(self, _event: MouseEvent):
        cords = self._coordinates()
        if cords is None:
            return
        for image_info in self.image_info.values():
            if image_info.roi_info.roi is None:
                continue
            if not image_info.coords_in(cords):
                continue
            moved_coords = image_info.translated_coords(cords)
            if component := image_info.roi_info.roi[tuple(moved_coords)]:
                self.component_clicked.emit(component)

    def get_roi_view_parameters(self, image_info: ImageInfo):
        res = super().get_roi_view_parameters(image_info=image_info)

        if (
            self.settings.get_from_profile(f"{self.name}.image_state.show_label", LabelEnum.Show_results)
            != LabelEnum.Show_selected
        ):
            return res

        selected_components = set(self.settings.chosen_components())
        for num in list(res):
            if num not in selected_components:
                res[num] = [0, 0, 0, 0]
        return res

    def get_tool_tip_text(self) -> str:
        text = super().get_tool_tip_text()
        text_list = []
        for el in self.components:
            if self.settings.component_is_chosen(el):
                text_list.append(f"☑{el}")
            else:
                text_list.append(f"☐{el}")
        return " ".join(text_list) + "\n" + text if text else " ".join(text_list)



import os
from typing import Any, Optional

import numpy as np
from napari.types import FullLayerData

from PartSegImage import ChannelInfo, Image, ImageWriter
from PartSegImage.image import DEFAULT_SCALE_FACTOR


def napari_write_labels(path: str, data: Any, meta: dict) -> Optional[str]:
    ext = os.path.splitext(path)[1]
    if not isinstance(data, np.ndarray) or ext not in {".tiff", ".tif", ".TIFF", ".TIF"}:
        return None
    scale_shift = min(data.ndim, 3)
    image = Image(
        data,
        spacing=np.divide(meta["scale"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        axes_order="TZYX"[-data.ndim :],
        channel_info=[ChannelInfo(name=meta["name"])],
        shift=np.divide(meta["translate"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        name="ROI",
    )
    ImageWriter.save(image, path)
    return path


def napari_write_images(path: str, layer_data: list[FullLayerData]) -> list[str]:
    ext = os.path.splitext(path)[1]
    base_shape = layer_data[0][0].shape
    if not all(isinstance(x[0], np.ndarray) and x[0].shape == base_shape for x in layer_data) or ext.lower() not in {
        ".tiff",
        ".tif",
    }:
        return []
    scale_shift = min(len(base_shape), 3)
    axes = "TZXY"
    channel_names = [meta["name"] for _, meta, _ in layer_data]
    meta = layer_data[0][1]
    if len(layer_data) == 1:
        data = layer_data[0][0]
        if data.shape[-1] < 6:
            axes += "C"
            scale_shift -= 1
            channel_names = [f'{meta["name"]} {i}' for i in range(1, data.shape[-1] + 1)]
        axes = axes[-data.ndim :]
    else:
        data = [x[0] for x in layer_data]
        axes = f"C{axes[-len(data[0].shape):]}"
        scale_shift -= 1
    image = Image(
        data,
        spacing=np.divide(meta["scale"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        axes_order=axes,
        channel_info=[ChannelInfo(name=x) for x in channel_names],
        shift=np.divide(meta["translate"], DEFAULT_SCALE_FACTOR)[-scale_shift:],
        name="Image",
    )
    ImageWriter.save(image, path)
    return [path]



import codecs
import contextlib
import os
import re

from setuptools import setup

current_dir = os.path.dirname(os.path.abspath(__file__))
package_dir = os.path.join(current_dir, "package")


def read(*parts):
    with codecs.open(os.path.join(current_dir, *parts)) as fp:
        return fp.read()


def readme():
    this_directory = os.path.abspath(os.path.dirname(__file__))
    reg = re.compile(r"(!\[[^]]*\])\((images/[^)]*)\)")
    reg2 = re.compile(r"releases/latest/download/PartSeg")
    with open(os.path.join(this_directory, "Readme.md"), encoding="utf8") as f:
        text = f.read()
        text = reg.sub(r"\1(https://raw.githubusercontent.com/4DNucleome/PartSeg/master/\2)", text)
        with contextlib.suppress(ImportError):
            from setuptools_scm import get_version

            text = reg2.sub(f"releases/download/v{get_version()}/PartSeg-{get_version()}", text)

    with open(os.path.join(this_directory, "changelog.md"), encoding="utf8") as f:
        chg = f.read()
        text += "\n\n" + chg.replace("# ", "## ")
    return text


changelog_path = os.path.join(os.path.dirname(__file__), "changelog.md")
changelog_result_path = os.path.join(os.path.dirname(__file__), "package", "PartSeg", "changelog.py")
if os.path.exists(changelog_path):
    with open(changelog_path, encoding="utf8") as ff:
        changelog_str = ff.read()
    with open(changelog_result_path, "w", encoding="utf8") as ff:
        ff.write(f'changelog = """\n{changelog_str}"""\n')


setup(
    include_package_data=True,
    long_description=readme(),
    long_description_content_type="text/markdown",
)



import os
from contextlib import suppress

from qtpy.QtCore import QByteArray, Qt
from qtpy.QtGui import QIcon, QKeyEvent, QKeySequence, QResizeEvent
from qtpy.QtWidgets import (
    QCheckBox,
    QComboBox,
    QGridLayout,
    QHBoxLayout,
    QInputDialog,
    QLabel,
    QMessageBox,
    QPushButton,
    QVBoxLayout,
    QWidget,
)
from superqt import ensure_main_thread

import PartSegData
from PartSeg import state_store
from PartSeg._roi_analysis.advanced_window import SegAdvancedWindow
from PartSeg._roi_analysis.batch_window import BatchWindow
from PartSeg._roi_analysis.calculation_pipeline_thread import CalculatePipelineThread
from PartSeg._roi_analysis.image_view import CompareImageView, ResultImageView, SynchronizeView
from PartSeg._roi_analysis.measurement_widget import MeasurementWidget
from PartSeg._roi_analysis.partseg_settings import PartSettings
from PartSeg.common_backend.base_settings import IO_SAVE_DIRECTORY
from PartSeg.common_backend.except_hook import show_warning
from PartSeg.common_gui.algorithms_description import AlgorithmChoose, InteractiveAlgorithmSettingsWidget
from PartSeg.common_gui.channel_control import ChannelProperty
from PartSeg.common_gui.custom_load_dialog import PLoadDialog
from PartSeg.common_gui.custom_save_dialog import PSaveDialog
from PartSeg.common_gui.equal_column_layout import EqualColumnLayout
from PartSeg.common_gui.exception_hooks import OPEN_ERROR, load_data_exception_hook
from PartSeg.common_gui.main_window import OPEN_DIRECTORY, OPEN_FILE, OPEN_FILE_FILTER, BaseMainMenu, BaseMainWindow
from PartSeg.common_gui.mask_widget import MaskDialogBase
from PartSeg.common_gui.multiple_file_widget import MultipleFileWidget
from PartSeg.common_gui.searchable_combo_box import SearchComboBox
from PartSeg.common_gui.stack_image_view import ColorBar
from PartSeg.common_gui.stacked_widget_with_selector import StackedWidgetWithSelector
from PartSeg.common_gui.universal_gui_part import TextShow
from PartSeg.common_gui.waiting_dialog import ExecuteFunctionDialog, WaitingDialog
from PartSegCore.algorithm_describe_base import ROIExtractionProfile
from PartSegCore.analysis import ProjectTuple, algorithm_description, load_functions
from PartSegCore.analysis.analysis_utils import SegmentationPipeline, SegmentationPipelineElement
from PartSegCore.analysis.io_utils import create_history_element_from_project
from PartSegCore.analysis.save_functions import save_dict
from PartSegCore.io_utils import WrongFileTypeException
from PartSegCore.project_info import HistoryElement, calculate_mask_from_project
from PartSegCore.roi_info import ROIInfo
from PartSegCore.segmentation.algorithm_base import ROIExtractionResult
from PartSegCore.utils import EventedDict
from PartSegImage import TiffImageReader

CONFIG_FOLDER = os.path.join(state_store.save_folder, "analysis")


class Options(QWidget):
    def __init__(
        self,
        settings: PartSettings,
        channel_control2: ChannelProperty,
        left_image: ResultImageView,
        synchronize: SynchronizeView,
    ):
        super().__init__()
        self._settings = settings
        self.left_panel = left_image
        self._ch_control2 = channel_control2
        self.synchronize_val = False
        self.hide_left_panel_chk = QCheckBox("Hide left panel")
        self.hide_left_panel_chk.stateChanged.connect(self.hide_left_panel)
        self.synchronize_checkbox = QCheckBox("Synchronize view")
        self.synchronize_checkbox.stateChanged.connect(synchronize.set_synchronize)
        self.interactive_use = QCheckBox("Interactive use")
        self.execute_btn = QPushButton("Execute")
        self.execute_btn.clicked.connect(self.execute_algorithm)
        self.execute_btn.setStyleSheet("QPushButton{font-weight: bold;}")
        self.save_pipe_btn = QPushButton("Save pipeline")
        self.save_pipe_btn.clicked.connect(self.save_pipeline)
        self.save_pipe_btn.setToolTip("Save current pipeline. Last element is last executed algorithm")
        self.choose_pipe = SearchComboBox()
        self.choose_pipe.addItem("<none>")
        self.choose_pipe.textActivated.connect(self.choose_pipeline)
        self.choose_pipe.setToolTip("Execute chosen pipeline")
        self.save_profile_btn = QPushButton("Save profile")
        self.save_profile_btn.setToolTip("Save values from current view")
        self.save_profile_btn.clicked.connect(self.save_profile)
        self.choose_profile = SearchComboBox()
        self.choose_profile.addItem("<none>")
        self.choose_profile.setToolTip("Select profile to restore its settings. Execute if interactive is checked")
        # image state
        self.compare_btn = QPushButton("Compare")
        self.compare_btn.setDisabled(True)
        self.compare_btn.clicked.connect(self.compare_action)
        left_image.hide_signal.connect(self.compare_btn.setHidden)

        self.update_tooltips()
        self.choose_profile.textActivated.connect(self.change_profile)
        self.interactive_use.stateChanged.connect(self.execute_btn.setDisabled)
        self.interactive_use.stateChanged.connect(self.interactive_change)
        self.algorithm_choose_widget = AlgorithmChoose(settings, algorithm_description.AnalysisAlgorithmSelection)
        self.algorithm_choose_widget.result.connect(self.execution_done)
        self.algorithm_choose_widget.finished.connect(self.calculation_finished)
        self.algorithm_choose_widget.value_changed.connect(self.interactive_algorithm_execute)
        self.algorithm_choose_widget.algorithm_changed.connect(self.interactive_algorithm_execute)
        self._settings.roi_profiles_changed.connect(self._update_profiles)
        self._settings.roi_pipelines_changed.connect(self._update_pipelines)
        self._update_pipelines()
        self._update_profiles()

        self.label = TextShow()

        self.setup_ui()

        settings.roi_changed.connect(self._refresh_compare_btn)
        settings.image_changed.connect(self._reset_compare_btn)

    def setup_ui(self):
        layout = QVBoxLayout()
        layout2 = QHBoxLayout()
        layout2.setSpacing(1)
        layout2.setContentsMargins(0, 0, 0, 0)
        layout3 = QHBoxLayout()
        layout3.setContentsMargins(0, 0, 0, 0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout5 = QHBoxLayout()
        layout5.setContentsMargins(0, 0, 0, 0)
        layout5.addWidget(self.save_pipe_btn)
        layout5.addWidget(self.choose_pipe)
        layout4 = QHBoxLayout()
        layout4.setContentsMargins(0, 0, 0, 0)
        layout4.addWidget(self.save_profile_btn)
        layout4.addWidget(self.choose_profile)
        layout3.addWidget(self.interactive_use)
        layout3.addWidget(self.execute_btn)
        layout.addLayout(layout5)
        layout.addLayout(layout4)
        layout.addLayout(layout3)
        layout.addWidget(self.algorithm_choose_widget, 1)
        layout.addWidget(self.label)
        layout2.addWidget(self.hide_left_panel_chk)
        layout2.addWidget(self.synchronize_checkbox)
        layout.addLayout(layout2)
        layout.addWidget(self._ch_control2)
        self.setLayout(layout)

    @ensure_main_thread
    def _update_profiles(self):
        self.update_combo_box(self.choose_profile, self._settings.roi_profiles)
        self.update_tooltips()

    @ensure_main_thread
    def _update_pipelines(self):
        self.update_combo_box(self.choose_pipe, self._settings.roi_pipelines)
        self.update_tooltips()

    def compare_action(self):
        if self.compare_btn.text() == "Compare":
            self._settings.set_segmentation_to_compare(self._settings.roi_info)
            self.compare_btn.setText("Remove")
        else:
            self._settings.set_segmentation_to_compare(ROIInfo(None))
            self.compare_btn.setText("Compare")

    def calculation_finished(self):
        self.execute_btn.setDisabled(self.interactive_use.isChecked())
        self.interactive_use.setEnabled(True)

    def save_pipeline(self):
        history = self._settings.get_history()
        if not history:
            QMessageBox.information(self, "No mask created", "There is no new mask created", QMessageBox.Ok)
            return
        mask_history = []
        for el in history:
            mask = el.mask_property
            segmentation = ROIExtractionProfile(
                name="Unknown",
                algorithm=el.roi_extraction_parameters["algorithm_name"],
                values=el.roi_extraction_parameters["values"],
            )
            new_el = SegmentationPipelineElement(mask_property=mask, segmentation=segmentation)
            mask_history.append(new_el)
        name = self._settings.last_executed_algorithm
        if not name:
            QMessageBox.information(self, "No segmentation", "No segmentation executed", QMessageBox.Ok)
            return
        values = self._settings.get_algorithm(f"algorithms.{name}", {})
        if isinstance(values, (dict, EventedDict)) and len(values) == 0:
            QMessageBox.information(self, "Some problem", "Please run execution again", QMessageBox.Ok)
            return
        current_segmentation = ROIExtractionProfile(name="Unknown", algorithm=name, values=values)

        while True:
            text, ok = QInputDialog.getText(self, "Pipeline name", "Input pipeline name here")
            if not ok:
                return
            if text in self._settings.roi_pipelines and QMessageBox.No == QMessageBox.warning(
                self,
                "Already exists",
                "Profile with this name already exist. Overwrite?",
                QMessageBox.Yes | QMessageBox.No,
                QMessageBox.No,
            ):
                continue
            profile = SegmentationPipeline(name=text, segmentation=current_segmentation, mask_history=mask_history)
            self._settings.roi_pipelines[text] = profile
            self._settings.dump()
            break

    def choose_pipeline(self, text):
        if text == "<none>":
            return
        pipeline = self._settings.roi_pipelines[text]
        process_thread = CalculatePipelineThread(self._settings.image, self._settings.mask, pipeline)
        dial = WaitingDialog(process_thread)

        if dial.exec_() and process_thread.result:
            pipeline_result = process_thread.result
            self._settings.mask = pipeline_result.mask
            self._settings.roi = pipeline_result.roi_info.roi
            self._settings.set_history(pipeline_result.history)
            self.label.setText(pipeline_result.description)
            self.algorithm_choose_widget.change_algorithm(pipeline.segmentation.algorithm, pipeline.segmentation.values)
        self.choose_pipe.setCurrentIndex(0)

    def update_tooltips(self):
        for i in range(1, self.choose_profile.count()):
            if self.choose_profile.itemData(i, Qt.ItemDataRole.ToolTipRole) is not None:
                continue
            text = self.choose_profile.itemText(i)
            profile: ROIExtractionProfile = self._settings.roi_profiles[text]
            tool_tip_text = str(profile)
            self.choose_profile.setItemData(i, tool_tip_text, Qt.ItemDataRole.ToolTipRole)
        for i in range(1, self.choose_pipe.count()):
            if self.choose_pipe.itemData(i, Qt.ItemDataRole.ToolTipRole) is not None:
                continue
            text = self.choose_pipe.itemText(i)
            profile: SegmentationPipeline = self._settings.roi_pipelines[text]
            tool_tip_text = str(profile)
            self.choose_pipe.setItemData(i, tool_tip_text, Qt.ItemDataRole.ToolTipRole)

    @staticmethod
    def update_combo_box(combo_box: QComboBox, dkt: dict):
        current_names = set(dkt.keys())
        prev_names = {combo_box.itemText(i) for i in range(1, combo_box.count())}
        new_names = current_names - prev_names
        delete_names = prev_names - current_names
        if len(delete_names) > 0:
            i = 1
            while i < combo_box.count():
                if combo_box.itemText(i) in delete_names:
                    combo_box.removeItem(i)
                else:
                    i += 1
        if len(new_names) > 0:
            combo_box.addItems(sorted(new_names))

    def keyPressEvent(self, event: QKeyEvent):
        if (
            event.key() in [Qt.Key.Key_Enter, Qt.Key.Key_Return]
            and event.modifiers() == Qt.KeyboardModifier.ControlModifier
        ):
            self.execute_btn.click()

    def save_profile(self):
        widget: InteractiveAlgorithmSettingsWidget = self.algorithm_choose_widget.current_widget()
        while True:
            text, ok = QInputDialog.getText(self, "Profile Name", "Input profile name here")
            if not ok:
                return
            if text in self._settings.roi_profiles and QMessageBox.StandardButton.No == QMessageBox.warning(
                self,
                "Already exists",
                "Profile with this name already exist. Overwrite?",
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            ):
                continue
            resp = ROIExtractionProfile(name=text, algorithm=widget.name, values=widget.get_values())
            self._settings.roi_profiles[text] = resp
            self._settings.dump()
            break

    def change_profile(self, val):
        self.choose_profile.setToolTip("")
        if val == "<none>":
            return
        interactive = self.interactive_use.isChecked()
        self.interactive_use.setChecked(False)
        profile = self._settings.roi_profiles[val]
        self.algorithm_choose_widget.change_algorithm(profile.algorithm, profile.values)
        self.choose_profile.blockSignals(True)
        self.choose_profile.setCurrentIndex(0)
        self.choose_profile.blockSignals(False)
        self.interactive_use.setChecked(interactive)

    @property
    def segmentation(self):
        return self._settings.roi

    @property
    def interactive(self):
        return self.interactive_use.isChecked()

    def hide_left_panel(self, val):
        self._settings.set_in_profile("hide_left_panel", val)
        if val:
            self.synchronize_val = self.synchronize_checkbox.isChecked()
            self.synchronize_checkbox.setChecked(False)
        else:
            self.synchronize_checkbox.setChecked(self.synchronize_val)
        self.synchronize_checkbox.setDisabled(val)
        self.left_panel.parent().setHidden(val)

    def interactive_change(self, val):
        if val:
            self.execute_algorithm()

    def algorithm_change(self, val):
        self._settings.set("current_algorithm", val)
        if self.interactive:
            self.execute_algorithm()

    def interactive_algorithm_execute(self):
        if self.interactive:
            self.execute_algorithm()

    def execute_algorithm(self):
        widget: InteractiveAlgorithmSettingsWidget = self.algorithm_choose_widget.current_widget()
        if self._settings.image.is_time and not widget.algorithm.support_time():
            QMessageBox.information(
                self, "Not supported", "This algorithm do not support time data. You can convert it in image adjust"
            )
            return
        if self._settings.image.is_stack and not widget.algorithm.support_z():
            QMessageBox.information(
                self, "Not supported", "This algorithm do not support stack data. You can convert it in image adjust"
            )
            return
        self._settings.last_executed_algorithm = widget.name
        self.execute_btn.setDisabled(True)
        self.interactive_use.setDisabled(True)
        widget.execute()

    def execution_done(self, segmentation: ROIExtractionResult):
        if segmentation.info_text:
            QMessageBox.information(self, "Algorithm info", segmentation.info_text)
        self._settings.set_segmentation_result(segmentation)
        self.label.setText(self.sender().get_info_text())

    def _refresh_compare_btn(self):
        self.compare_btn.setEnabled(bool(self._settings.roi_info.bound_info))

    def _reset_compare_btn(self):
        self.compare_btn.setText("Compare")

    def showEvent(self, _event):
        self.hide_left_panel_chk.setChecked(self._settings.get_from_profile("hide_left_panel", False))


class MainMenu(BaseMainMenu):
    def __init__(self, settings: PartSettings, main_window):
        super().__init__(settings, main_window)
        self.settings = settings
        self.open_btn = QPushButton("Open")
        self.save_btn = QPushButton("Save")
        self.advanced_btn = QPushButton("Settings and Measurement")
        self.mask_manager_btn = QPushButton("Mask manager")
        self.batch_processing_btn = QPushButton("Batch Processing")

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 4, 4)
        layout.addWidget(self.open_btn)
        layout.addWidget(self.save_btn)
        layout.addWidget(self.advanced_btn)
        layout.addWidget(self.mask_manager_btn)
        layout.addWidget(self.batch_processing_btn)
        self.setLayout(layout)

        self.open_btn.clicked.connect(self.load_data)
        self.save_btn.clicked.connect(self.save_file)
        self.advanced_btn.clicked.connect(self.advanced_window_show)
        self.mask_manager_btn.clicked.connect(self.mask_manager)
        self.batch_processing_btn.clicked.connect(self.batch_window)
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    def resizeEvent(self, event: QResizeEvent):
        if event.size().width() < 800:
            self.batch_processing_btn.setVisible(False)
        else:
            self.batch_processing_btn.setVisible(True)

    def keyPressEvent(self, event: QKeyEvent):
        if event.matches(QKeySequence.Save):
            self.save_file()
        elif event.matches(QKeySequence.Open):
            self.load_data()
        super().keyPressEvent(event)

    def save_file(self):
        base_values = self.settings.get("save_parameters", {})
        dial = PSaveDialog(
            save_dict,
            system_widget=False,
            base_values=base_values,
            settings=self.settings,
            path=IO_SAVE_DIRECTORY,
            filter_path="io.save_filter",
        )
        dial.selectFile(os.path.splitext(os.path.basename(self.settings.image_path))[0])
        if dial.exec_():
            save_location, selected_filter, save_class, values = dial.get_result()
            project_info = self.settings.get_project_info()
            if save_class.need_segmentation() and project_info.roi_info.roi is None:
                QMessageBox.information(self, "No segmentation", "Cannot save without segmentation")
                return
            if save_class.need_mask() and project_info.mask is None:
                QMessageBox.information(self, "No mask", "Cannot save without mask")
                return
            base_values[selected_filter] = values

            def exception_hook(exception):  # pragma: no cover
                if isinstance(exception, ValueError):
                    show_warning("Save error", f"Error during saving\n{exception}", exception=exception)
                else:
                    raise exception

            dial2 = ExecuteFunctionDialog(
                save_class.save, [save_location, project_info, values], exception_hook=exception_hook
            )
            dial2.exec_()

    def mask_manager(self):
        if self.settings.roi is None:
            QMessageBox.information(self, "No segmentation", "Cannot create mask without segmentation")
            return
        dial = MaskDialog(self.settings)
        dial.exec_()

    def load_data(self):
        def exception_hook(exception):
            if isinstance(exception, WrongFileTypeException):
                show_warning(
                    OPEN_ERROR,
                    "No needed files inside archive. Most probably you choose file from segmentation mask",
                    exception=exception,
                )
            else:
                load_data_exception_hook(exception)

        try:
            dial = PLoadDialog(
                load_functions.load_dict, settings=self.settings, path=OPEN_DIRECTORY, filter_path=OPEN_FILE_FILTER
            )
            file_path = self.settings.get(OPEN_FILE, "")
            if os.path.isfile(file_path):
                dial.selectFile(file_path)
            if dial.exec_():
                result = dial.get_result()
                self.settings.set(OPEN_FILE, result.load_location[0])
                self.settings.add_last_files(result.load_location, result.load_class.get_name())
                dial2 = ExecuteFunctionDialog(
                    result.load_class.load,
                    [result.load_location],
                    {"metadata": {"default_spacing": self.settings.image_spacing}},
                    exception_hook=exception_hook,
                )
                if dial2.exec_():
                    result = dial2.get_result()
                    self.set_data(result)

        except ValueError as e:  # pragma: no cover
            show_warning("Open error", f"{e}", exception=e)

    def batch_window(self):
        if self.main_window.batch_window is None:
            self.main_window.batch_window = BatchWindow(self.settings)
            self.main_window.batch_window.show()

        elif self.main_window.batch_window.isVisible():
            self.main_window.batch_window.activateWindow()
        else:
            self.main_window.batch_window.show()

    def advanced_window_show(self):
        if self.main_window.advanced_window.isVisible():
            self.main_window.advanced_window.activateWindow()
        else:
            self.main_window.advanced_window.show()


class MaskDialog(MaskDialogBase):
    # FIXME add tests
    def next_mask(self):
        project_info: ProjectTuple = self.settings.get_project_info()
        mask_property = self.mask_widget.get_mask_property()
        self.settings.set("mask_manager.mask_property", mask_property)
        mask = calculate_mask_from_project(mask_description=mask_property, project=project_info)
        self.settings.add_history_element(
            create_history_element_from_project(
                project_info,
                mask_property,
            )
        )
        if self.settings.history_redo_size():
            history: HistoryElement = self.settings.history_next_element()
            self.settings.set("current_algorithm", history.roi_extraction_parameters["algorithm_name"])
            self.settings.set(
                f"algorithm.{history.roi_extraction_parameters['algorithm_name']}",
                history.roi_extraction_parameters["values"],
            )
        self.settings.mask = mask
        self.close()

    def prev_mask(self):
        history: HistoryElement = self.settings.history_pop()
        algorithm_name = self.settings.last_executed_algorithm
        algorithm_values = self.settings.get_algorithm(f"algorithms.{algorithm_name}")
        self.settings.fix_history(algorithm_name=algorithm_name, algorithm_values=algorithm_values)
        self.settings.set("current_algorithm", history.roi_extraction_parameters["algorithm_name"])
        self.settings.set(
            f"algorithm.{history.roi_extraction_parameters['algorithm_name']}",
            history.roi_extraction_parameters["values"],
        )
        self.settings.roi, self.settings.mask = history.get_roi_info_and_mask()
        self.close()


class MainWindow(BaseMainWindow):
    settings: PartSettings

    @classmethod
    def get_setting_class(cls) -> type[PartSettings]:
        return PartSettings

    initial_image_path = PartSegData.segmentation_analysis_default_image

    def __init__(  # noqa: PLR0915
        self, config_folder=CONFIG_FOLDER, title="PartSeg", settings=None, signal_fun=None, initial_image=None
    ):
        super().__init__(config_folder, title, settings, load_functions.load_dict, signal_fun)
        self.channel_info = "result_image"
        self.files_num = 2
        self.setMinimumWidth(600)
        # this isinstance is only for type hinting in IDE
        self.main_menu = MainMenu(self.settings, self)
        self.channel_control2 = ChannelProperty(self.settings, start_name="result_image")
        self.raw_image = CompareImageView(self.settings, self.channel_control2, "raw_image")
        self.measurements = MeasurementWidget(self.settings)
        self.left_stack = StackedWidgetWithSelector()
        self.left_stack.addWidget(self.raw_image, "Image")
        self.left_stack.addWidget(self.measurements, "Measurements")
        self.result_image = ResultImageView(self.settings, self.channel_control2, "result_image")
        self.color_bar = ColorBar(self.settings, [self.raw_image, self.result_image])
        self.info_text = QLabel()
        self.info_text.setMinimumHeight(25)
        self.raw_image.text_info_change.connect(self.info_text.setText)
        self.result_image.text_info_change.connect(self.info_text.setText)
        self.synchronize_tool = SynchronizeView(self.raw_image, self.result_image, self)
        self.options_panel = Options(self.settings, self.channel_control2, self.raw_image, self.synchronize_tool)
        self.settings.image_changed.connect(self.image_read)
        self.advanced_window = SegAdvancedWindow(self.settings, reload_list=[self.reload])
        self.batch_window = None  # BatchWindow(self.settings)

        self.multiple_files = MultipleFileWidget(self.settings, load_functions.load_dict, True)
        self.multiple_files.setVisible(self.settings.get("multiple_files_widget", False))

        if initial_image is None:
            reader = TiffImageReader()
            im = reader.read(self.initial_image_path)
            im.file_path = ""
            self.settings.image = im
        elif initial_image is not False:
            self.settings.image = initial_image

        self._setup_menu_bar()

        icon = QIcon(os.path.join(PartSegData.icons_dir, "icon.png"))
        self.setWindowIcon(icon)

        layout = QGridLayout()
        layout.setSpacing(0)
        info_layout = QHBoxLayout()
        info_layout.addWidget(self.left_stack.selector)
        info_layout.addWidget(self.options_panel.compare_btn)
        info_layout.addWidget(self.info_text, 1, Qt.AlignmentFlag.AlignHCenter)

        image_layout = EqualColumnLayout()
        image_layout.addWidget(self.left_stack)
        image_layout.addWidget(self.result_image)

        layout.setSpacing(0)
        layout.addWidget(self.main_menu, 0, 0, 1, 3)
        layout.addLayout(info_layout, 3, 1, 1, 2)
        layout.addWidget(self.multiple_files, 2, 0)
        layout.addWidget(self.color_bar, 2, 1)
        layout.addLayout(image_layout, 2, 2, 1, 1)
        layout.addWidget(self.options_panel, 0, 3, 3, 1)
        layout.setColumnStretch(2, 1)
        widget = QWidget()
        widget.setLayout(layout)
        self.setCentralWidget(widget)
        with suppress(KeyError):
            geometry = self.settings.get_from_profile("main_window_geometry")
            self.restoreGeometry(QByteArray.fromHex(bytes(geometry, "ascii")))

    def _setup_menu_bar(self):
        menu_bar = self.menuBar()
        file_menu = menu_bar.addMenu("File")
        file_menu.addAction("&Open").triggered.connect(self.main_menu.load_data)
        file_menu.addMenu(self.recent_file_menu)
        file_menu.addAction("&Save").triggered.connect(self.main_menu.save_file)
        file_menu.addAction("Batch processing").triggered.connect(self.main_menu.batch_window)
        view_menu = menu_bar.addMenu("View")
        view_menu.addAction("Settings and Measurement").triggered.connect(self.main_menu.advanced_window_show)
        view_menu.addAction("Additional output").triggered.connect(self.additional_layers_show)
        view_menu.addAction("Additional output with data").triggered.connect(lambda: self.additional_layers_show(True))
        view_menu.addAction("Napari viewer").triggered.connect(self.napari_viewer_show)
        view_menu.addAction("Toggle Multiple Files").triggered.connect(self.toggle_multiple_files)
        view_menu.addAction("Toggle left panel").triggered.connect(self.toggle_left_panel)
        view_menu.addAction("Toggle console").triggered.connect(self._toggle_console)
        view_menu.addAction("Toggle scale bar").triggered.connect(self._toggle_scale_bar)
        action = view_menu.addAction("Screenshot right panel")
        action.triggered.connect(self.screenshot(self.result_image))
        action.setShortcut(QKeySequence.Print)
        view_menu.addAction("Screenshot left panel").triggered.connect(self.screenshot(self.raw_image))
        image_menu = menu_bar.addMenu("Image operations")
        image_menu.addAction("Image adjustment").triggered.connect(self.image_adjust_exec)
        image_menu.addAction("Mask manager").triggered.connect(self.main_menu.mask_manager)
        help_menu = menu_bar.addMenu("Help")
        help_menu.addAction("State directory").triggered.connect(self.show_settings_directory)
        help_menu.addAction("About").triggered.connect(self.show_about_dialog)

    def _toggle_scale_bar(self):
        self.raw_image.toggle_scale_bar()
        self.result_image.toggle_scale_bar()
        super()._toggle_scale_bar()

    def toggle_left_panel(self):
        self.options_panel.hide_left_panel(not self.settings.get_from_profile("hide_left_panel"))

    def image_read(self):
        super().image_read()
        self.options_panel.interactive_algorithm_execute()

    def reload(self):
        self.options_panel.algorithm_choose_widget.reload(algorithm_description.AnalysisAlgorithmSelection)

    def closeEvent(self, event):
        self.settings.set_in_profile("main_window_geometry", self.saveGeometry().toHex().data().decode("ascii"))
        self.options_panel.algorithm_choose_widget.recursive_get_values()
        if self.batch_window is not None:
            if self.batch_window.is_working():
                ret = QMessageBox.warning(
                    self,
                    "Batch work",
                    "Batch work is not finished. Would you like to terminate it?",
                    QMessageBox.No | QMessageBox.Yes,
                )
                if ret == QMessageBox.Yes:
                    self.batch_window.terminate()
                else:
                    event.ignore()
                    return
            self.batch_window.close()
        self.advanced_window.close()
        del self.batch_window
        del self.advanced_window
        super().closeEvent(event)

    @staticmethod
    def get_project_info(file_path, image, roi_info=None):
        if roi_info is None:
            roi_info = ROIInfo(None)
        return ProjectTuple(file_path=file_path, image=image, roi_info=roi_info)

    def set_data(self, data):
        self.main_menu.set_data(data)

    def change_theme(self, event):
        self.raw_image.set_theme(self.settings.theme_name)
        self.result_image.set_theme(self.settings.theme_name)
        super().change_theme(event)



"""
This module is designed to provide more stable api for PartSeg plugins.
All operations ar class based. If, in base object, some function is @classmethod, the it can be called
without creating class instance. So also plugins need to satisfy it.

RegisterEnum: used as key in all other members. Elements are described in its docstring

base_class_dict: Dict in which information about base class for given operation is stored.
The inheritance is not checked, but api must be implemented.

register: function for registering operation in inner structures.

register_dict: holds information where register given operation type. Strongly suggest to use register function instead.
"""

from enum import Enum

from PartSegCore import io_utils
from PartSegCore.algorithm_describe_base import AlgorithmDescribeBase
from PartSegCore.analysis import algorithm_description as analysis_algorithm_description
from PartSegCore.analysis import load_functions, measurement_base, measurement_calculation, save_functions
from PartSegCore.image_transforming import TransformBase, image_transform_dict
from PartSegCore.mask import algorithm_description as mask_algorithm_description
from PartSegCore.mask import io_functions
from PartSegCore.segmentation import (
    noise_filtering,
    restartable_segmentation_algorithms,
    segmentation_algorithm,
    threshold,
    watershed,
)

qss_list = []


class RegisterEnum(Enum):
    """
    Given types of operation are supported as plugins
    """

    flow = 0  #: algorithm for calculation flow from core object to borders. For spiting touching objects, deprecated
    threshold = 1  #: threshold algorithms. From greyscale array to binary array
    noise_filtering = 2  #: filter noise from image
    analysis_algorithm = 3  #: algorithm for creating segmentation in analysis PartSeg part
    mask_algorithm = 4  #: algorithm for creating segmentation in mask PartSeg part
    analysis_save = 5  #: save functions for analysis part
    analysis_load = 6  #: load functions for analysis part
    mask_load = 7  #: load functions for mask part
    image_transform = 8  #: transform image, like interpolation
    mask_save_parameters = 9  #: save metadata for mask part (currently creating json file)
    mask_save_components = 10  #: save each segmentation component in separate file. Save location is directory
    mask_save_segmentation = 11  #: save project (to one file) in mask part
    analysis_measurement = 12  #: measurements algorithms (analysis mode)
    roi_analysis_segmentation_algorithm = 13  #: algorithm for creating segmentation in analysis PartSeg part
    roi_mask_segmentation_algorithm = 14  #: algorithm for creating segmentation in mask PartSeg part
    _qss_register = 15  #: new qss styles
    sprawl = 16  #: algorithm for calculation flow from core object to borders. For spiting touching objects, deprecated


# noinspection DuplicatedCode
register_dict = {
    RegisterEnum.flow: watershed.WatershedSelection.__register__,
    RegisterEnum.sprawl: watershed.WatershedSelection.__register__,
    RegisterEnum.threshold: threshold.ThresholdSelection.__register__,
    RegisterEnum.noise_filtering: noise_filtering.NoiseFilterSelection.__register__,
    RegisterEnum.analysis_algorithm: analysis_algorithm_description.AnalysisAlgorithmSelection.__register__,
    RegisterEnum.mask_algorithm: mask_algorithm_description.MaskAlgorithmSelection.__register__,
    RegisterEnum.analysis_save: save_functions.save_dict,
    RegisterEnum.analysis_load: load_functions.load_dict,
    RegisterEnum.mask_load: io_functions.load_dict,
    RegisterEnum.image_transform: image_transform_dict,
    RegisterEnum.mask_save_parameters: io_functions.save_parameters_dict,
    RegisterEnum.mask_save_components: io_functions.save_components_dict,
    RegisterEnum.mask_save_segmentation: io_functions.save_segmentation_dict,
    RegisterEnum.analysis_measurement: measurement_calculation.MEASUREMENT_DICT,
    RegisterEnum.roi_analysis_segmentation_algorithm: analysis_algorithm_description.AnalysisAlgorithmSelection.__register__,  # noqa: E501
    RegisterEnum.roi_mask_segmentation_algorithm: mask_algorithm_description.MaskAlgorithmSelection.__register__,
}

# noinspection DuplicatedCode
base_class_dict = {
    RegisterEnum.sprawl: watershed.BaseWatershed,
    RegisterEnum.flow: watershed.BaseWatershed,
    RegisterEnum.threshold: threshold.BaseThreshold,
    RegisterEnum.noise_filtering: noise_filtering.NoiseFilteringBase,
    RegisterEnum.analysis_algorithm: restartable_segmentation_algorithms.RestartableAlgorithm,
    RegisterEnum.mask_algorithm: segmentation_algorithm.ROIExtractionAlgorithm,
    RegisterEnum.analysis_save: io_utils.SaveBase,
    RegisterEnum.analysis_load: io_utils.LoadBase,
    RegisterEnum.mask_load: io_utils.LoadBase,
    RegisterEnum.image_transform: TransformBase,
    RegisterEnum.mask_save_parameters: io_utils.SaveBase,
    RegisterEnum.mask_save_components: io_utils.SaveBase,
    RegisterEnum.mask_save_segmentation: io_utils.SaveBase,
    RegisterEnum.analysis_measurement: measurement_base.MeasurementMethodBase,
}  # dict with base class for given type of algorithm, keys are :py:class:`RegisterEnum`

reload_module_list = [
    threshold,
    watershed,
    segmentation_algorithm,
    restartable_segmentation_algorithms,
    noise_filtering,
    io_functions,
    mask_algorithm_description,
    analysis_algorithm_description,
    measurement_calculation,
    save_functions,
    load_functions,
]


def register(target: type[AlgorithmDescribeBase], target_type: RegisterEnum, replace=False, old_names=None):
    """
    Function for registering new operations in PartSeg inner structures.

    :param target: operation to register in PartSeg inner structures
    :param target_type: Which type of operation.
    :param replace: force to replace operation if same name is defined. Dangerous.
    """
    if target_type == RegisterEnum._qss_register:  # pylint: disable=protected-access
        qss_list.append(target)
    else:
        register_dict[target_type].register(target, replace=replace, old_names=old_names)
